import json

def sum_numbers(ussd_request):
    return int(ussd_request.session['first_number']) + \
           int(ussd_request.session['second_number'])


class MockResponse:
    def __init__(self, json_data, status=200):
        self.json_data = json_data
        self.status_code = status
        self.content = json.dumps(json_data).encode()

    def json(self):
        return self.json_data



from ussd.tests import UssdTestCase


class TestInputHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        enter_height={
            "validators": {
                0: {'text': ['This field is required.']}
            },
            "next_screen": ['This field is required.']
        },
        enter_age={
            "input_identifier": ['This field is required.'],
            "next_screen": ['thank_you_screen is missing in ussd journey'],
            "options": ['Not a valid list.']
        },
        show_information={
            "type": ['Invalid screen type not supported']
        },
        hidden_fields={
            "initial_screen": ["This field is required."]
        },
        invalid_next_screen=dict(
            next_screen={'next_screen': ['This field is required.']}
        )
    )

    def test_showing_screen_content(self):
        ussd_client = self.ussd_client()

        self.assertEqual(
            "Enter your height\n",
            ussd_client.send('')  # dial in
        )

        self.assertEqual(
            "Enter your age\n1. back\n",
            ussd_client.send('5')  # enter height
        )

        # confirm height entered is 5 before going back
        self.assertEqual(
            self.ussd_session(ussd_client.session_id)["height"],
            '5'
        )

        self.assertEqual(
            "Enter your height\n",
            ussd_client.send('1')  # enter one to go back
        )

        ussd_client.send('6')  # enter height

        self.assertEqual(
            "Your age is 24 and your height is 6.\n"
            "Enter anything to go back to the first screen\n",
            ussd_client.send('24')  # enter age
        )

    def test_multilanguage_support(self):
        ussd_client = self.ussd_client(language='sw')

        # Dial in
        response = ussd_client.send('1')

        self.assertEqual(
            "Weka ukubwa lako\n",
            response
        )

        response = ussd_client.send('7')

        self.assertEqual(
            "Weka miaka yako\n1. rudi\n",
            response
        )

        response = ussd_client.send('23')

        self.assertEqual(
            "Miaka yako in 23 na ukubwa wako in 7.\n"
            "Weka kitu ingine yoyote unende "
            "kwenye screen ya kwanza\n",
            response
        )

    def test_input_validation(self):
        ussd_client = self.ussd_client()

        # dial in
        ussd_client.send('')

        # enter invalid height
        response = ussd_client.send('mwas')

        # should get a invalid error message
        self.assertEqual(
            "Enter number between 1 and 7\n",
            response
        )

        # enter valid height
        response = ussd_client.send('6')

        self.assertEqual(
            "Enter your age\n1. back\n",
            response
        )

        # enter invalid age greater thatn 100
        response = ussd_client.send('150')

        self.assertEqual(
            "Number over 100 is not allowed\n",
            response
        )

        # enter a valid age
        response = ussd_client.send('23')

        self.assertEqual(
            "Your age is 23 and your height is 6.\n"
            "Enter anything to go back to the first screen\n",
            response
        )

    def test_next_screen_configuration(self):
        ussd_client = self.ussd_client()
        ussd_client.send('')  # dial in

        self.assertEqual(
            "We are not interested with height above 60",
            ussd_client.send('60')
        )

        ussd_client = self.ussd_client()
        ussd_client.send('')  # dial in

        self.assertEqual(
            "We are not interested with height below 30",
            ussd_client.send('30')
        )



import json
import os
import uuid


from unittest import TestCase
from ussd.core import render_journey_as_graph, render_journey_as_mermaid_text, UssdEngine, UssdRequest, UssdResponse
from ussd.tests.sample_screen_definition import path
from ussd.store.journey_store.YamlJourneyStore import YamlJourneyStore
from ussd.session_store import SessionStore
from simplekv.fs import FilesystemStore


class UssdTestCase(object):
    """
    this contains two test that are required in each screen test case
    """

    class BaseUssdTestCase(TestCase):
        validate_ussd = True

        session_store = FilesystemStore("./session_data")

        def setUp(self):
            self.journey_store = YamlJourneyStore("/usr/src/app/ussd/tests/sample_screen_definition")

            file_prefix = self.__module__.split('.')[-1].replace('test_', '')
            self.journey_name = file_prefix
            journey_version_suffix = file_prefix + "_conf"

            self.valid_version = 'valid_' + journey_version_suffix
            self.invalid_version = 'invalid_' + journey_version_suffix

            self.mermaid_file = path + file_prefix + '/' + '/' + 'valid_' + file_prefix + '_mermaid.txt'
            self.graph_file = path + '/' + file_prefix + '/' + 'valid_' + file_prefix + '_graph.json'
            self.namespace = self.__module__.split('.')[-1]
            self.maxDiff = None

            super(UssdTestCase.BaseUssdTestCase, self).setUp()

            #

        def _test_ussd_validation(self, version_to_validate, expected_validation,
                                  expected_errors):

            if self.validate_ussd:
                ussd_screens = self.journey_store.get(self.journey_name, version_to_validate)

                is_valid, error_message = UssdEngine.validate_ussd_journey(
                    ussd_screens)

                self.assertEqual(is_valid, expected_validation, error_message)

                for key, value in expected_errors.items():
                    args = (value, error_message[key], key)
                    if isinstance(value, dict):
                        self.assertDictEqual(*args)
                    else:
                        self.assertEqual(*args)

                self.assertDictEqual(error_message,
                                     expected_errors)

        def testing_valid_customer_journey(self):
            self._test_ussd_validation(self.valid_version, True, {})

        def testing_invalid_customer_journey(self):

            try:
                self._test_ussd_validation(self.invalid_version, False,
                                       getattr(self,
                                               "validation_error_message",
                                               {}))
            except Exception as e:
                if not (hasattr(self, "expected_error") and isinstance(e, self.__getattribute__("expected_error"))):
                    raise e

        def test_rendering_graph_js(self):
            if os.path.exists(self.graph_file):
                ussd_screens = self.journey_store.get(self.journey_name, self.valid_version)

                actual_graph_js = render_journey_as_graph(ussd_screens)

                expected_graph_js = json.loads(self.read_file_content(self.graph_file))

                for key, value in expected_graph_js["vertices"].items():
                    if value.get('id') == 'test_explicit_dict_loop':
                        for i in (
                                "a for apple\n",
                                "b for boy\n",
                                "c for cat\n"
                        ):
                            self.assertRegex(value.get('text'), i)
                    else:
                        self.assertDictEqual(value, actual_graph_js.vertices[key])
                # self.assertDictEqual(expected_graph_js["vertices"], actual_graph_js.vertices)

                for index, value in enumerate(expected_graph_js['edges']):
                    self.assertDictEqual(value, actual_graph_js.get_edges()[index])
                self.assertEqual(expected_graph_js["edges"], actual_graph_js.get_edges())

        def test_rendering_mermaid_js(self):
            if os.path.exists(self.mermaid_file):
                ussd_screens = self.journey_store.get(self.journey_name, self.valid_version)

                mermaid_text_format = render_journey_as_mermaid_text(ussd_screens)

                file_content = self.read_file_content(self.mermaid_file)

                expected_text_lines = file_content.split('\n')
                actual_text_lines = mermaid_text_format.split('\n')

                for index, line in enumerate(expected_text_lines):
                    self.assertEqual(line, actual_text_lines[index])

                self.assertEqual(mermaid_text_format, file_content)

        def read_file_content(self, file_path):
            with open(file_path) as f:
                mermaid_text = f.read()
            return mermaid_text

        def ussd_session(self, session_id):
            return SessionStore(session_id, kv_store=self.session_store)

        def ussd_client(self, generate_customer_journey=True, **kwargs):
            class UssdTestClient(object):
                def __init__(self, session_id=None, phone_number=200,
                             language='en', extra_payload=None, ):

                    if extra_payload is None:
                        extra_payload = {}
                    self.phone_number = phone_number
                    self.language = language
                    self.session_id = session_id \
                        if session_id is not None \
                        else str(uuid.uuid4())
                    self.extra_payload = extra_payload

                def send(self, ussd_input, raw=False):
                    payload = {
                        "session_id": self.session_id,
                        "ussd_input": ussd_input,
                        "phone_number": self.phone_number,
                        "language": self.language,
                    }
                    payload.update(self.extra_payload)

                    ussd_request = UssdRequest(**payload)

                    response = UssdEngine(ussd_request).ussd_dispatcher()

                    if raw:
                        return response
                    return str(response)

            customer_journey_conf = {
                'journey_name': self.journey_name,
                'journey_version': self.valid_version,
                "journey_store": self.journey_store
            }

            if kwargs.get('extra_payload'):
                customer_journey_conf.update(kwargs['extra_payload'])

            kwargs['extra_payload'] = customer_journey_conf

            return UssdTestClient(**kwargs)



import sys

PY36 = sys.version_info >= (3, 6)


class cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    A cached property can be made out of an existing method:
    (e.g. ``url = cached_property(get_absolute_url)``).
    On Python < 3.6, the optional ``name`` argument must be provided, e.g.
    ``url = cached_property(get_absolute_url, name='url')``.
    """
    name = None

    @staticmethod
    def func(instance):
        raise TypeError(
            'Cannot use cached_property instance without calling '
            '__set_name__() on it.'
        )

    @staticmethod
    def _is_mangled(name):
        return name.startswith('__') and not name.endswith('__')

    def __init__(self, func, name=None):
        if PY36:
            self.real_func = func
        else:
            func_name = func.__name__
            name = name or func_name
            if not (isinstance(name, str) and name.isidentifier()):
                raise ValueError(
                    "%r can't be used as the name of a cached_property." % name,
                )
            if self._is_mangled(name):
                raise ValueError(
                    'cached_property does not work with mangled methods on '
                    'Python < 3.6 without the appropriate `name` argument. See '
                    'https://docs.djangoproject.com/en/%s/ref/utils/'
                    '#cached-property-mangled-name' % 1,
                )
            self.name = name
            self.func = func
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name
            self.func = self.real_func
        elif name != self.name:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                "(%r and %r)." % (self.name, name)
            )

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res



from unittest import TestCase
from ussd.store.journey_store import JourneyStoreApi
from unittest import mock
from copy import deepcopy
from marshmallow import ValidationError

b = 0


def helper(a, name="mwas"):
    global b
    b += 1


class BaseTestCase(TestCase):
    maxDiff = None

    journey = {
        "initial_screen": {
            "type": "initial_screen",
            "next_screen": "end_screen",
            "default_language": "en"
        },
        "end_screen": {
            "type": "quit_screen",
            "text": "end screen"
        }
    }

    driver_config = dict(
        journey_directory="./.journeys"
    )

    @staticmethod
    def _get_store_config_without_driver_name(store_config):
        # bad idea here but its only in tests
        # won't affect production performance
        new_store_config = deepcopy(store_config)

        if new_store_config.get("driver"):
            del new_store_config["driver"]
        return new_store_config

    @mock.patch("ussd.store.journey_store.JourneyStore.save")
    @mock.patch("ussd.store.journey_store.YamlJourneyStore.YamlJourneyStore.__init__")
    @mock.patch("ussd.store.journey_store.JourneyStoreApi._parameter_validation")
    def test_get_attribute(self, journey_store_parameter_validation, journey_store_init_mock, journey_store_save_mock):
        journey_store_parameter_validation.return_value = {}
        journey_store_init_mock.return_value = None

        JourneyStoreApi(driver_config=self.driver_config).save("test_save", self.journey, random_key_word="random_key_word")

        journey_store_parameter_validation.assert_called_once_with(
            journey_store_save_mock, "test_save", self.journey, random_key_word="random_key_word"
        )

        journey_store_init_mock.assert_called_once_with(
            **self._get_store_config_without_driver_name(self.driver_config),
        )

        journey_store_save_mock.assert_called_once_with(
            "test_save", self.journey, random_key_word="random_key_word"
        )

    def test_parameter_validation(self):

        # calling helper with only one argument
        self.assertDictEqual({}, JourneyStoreApi._parameter_validation(helper, a="a"))

        # call with a keyword
        # calling helper with only one argument and keyword
        self.assertDictEqual({}, JourneyStoreApi._parameter_validation(helper, "b", name="mwaside"))

        # calling with all keywords
        self.assertDictEqual({}, JourneyStoreApi._parameter_validation(helper, a="b", name="mwaside"))

        # calling with incorrect keyword
        self.assertDictEqual(
            dict(
                mwaside=["This field is not required"]
            ),
            JourneyStoreApi._parameter_validation(helper, "b", mwaside="mwaside")
        )

        # calling with incorrect arg
        self.assertDictEqual(
            dict(
                a=["This field is required"]
            ),
            JourneyStoreApi._parameter_validation(helper, name="mwaside")
        )

        # check it was not called
        helper('a')
        self.assertEqual(b, 1)

    def test_crud(self):
        store = JourneyStoreApi(driver_config=self.driver_config)

        name = "foo"
        version = "0.0.0"

        # save journey
        store.save(name=name, journey=self.journey, version=version)

        journey = store.get(name=name, version=version)
        self.assertEqual(self.journey, journey)

        store.delete(name=name)

        # confirm journey has been deleted
        self.assertRaises(ValidationError, store.get, name=name, version=version)

    def test_handle_action(self):

        store = JourneyStoreApi(driver_config=self.driver_config)

        name = "foo"
        version = "0.0.1"

        # save journey
        store.handle_action(action="save", name=name, journey=self.journey, version=version)

        # get journey that was created
        journey = store.handle_action(action="get", name=name, version=version)
        self.assertEqual(self.journey, journey)

        # create another journey
        journey2 = deepcopy(self.journey)
        journey2["end_screen"]["text"] = "journey 2"
        store.handle_action(action="save", name="bar", journey=journey2, version=version)
        # get journey that was created
        journey = store.handle_action(action="get", name="bar", version=version)
        self.assertEqual(journey2, journey)

        # now get all journeys
        journeys = store.handle_action(action="get")
        self.assertDictEqual({"foo": {version: self.journey}, "bar": {version: journey2}}, journeys)

        # delete journey that was created
        store.handle_action(action="delete", name=name)

        # confirm journey has been deleted
        self.assertRaises(ValidationError, store.get, name=name, version=version)

        # validate it action should be provided
        with self.assertRaises(ValidationError) as validation_err:
            # get journey without specifying the action
            store.handle_action(name=name, version=version)

        self.assertDictEqual(
            dict(
                action=["This field is required"]
            ),
            validation_err.exception.normalized_messages()
        )

        # test with invalid action
        with self.assertRaises(ValidationError) as validation_err:
            # get journey with invalid action
            store.handle_action(action="does_not_exist", name=name, version=version)

        self.assertDictEqual(
            dict(
                action=["action 'does_not_exist' is not allowed, only this methods are allowed; "
                        "save, delete, get, all, flush"]
            ),
            validation_err.exception.normalized_messages()
        )

    @mock.patch("ussd.store.journey_store.JourneyStoreApi._parameter_validation")
    @mock.patch("ussd.store.journey_store.JourneyStore.save")
    def test_post_alias_to_save(self, journey_store_save_mock, journey_store_parameter_validation):
        journey_store_parameter_validation.return_value = {}

        JourneyStoreApi(driver_config=self.driver_config).handle_action(action="post", name="test_save", journey=self.journey)

        journey_store_save_mock.assert_called_once_with(
            name="test_save", journey=self.journey
        )

    @mock.patch("ussd.store.journey_store.JourneyStoreApi._parameter_validation")
    @mock.patch("ussd.store.journey_store.JourneyStore.save")
    def test_put_alias_to_save(self, journey_store_save_mock, journey_store_parameter_validation):
        journey_store_parameter_validation.return_value = {}

        JourneyStoreApi(driver_config=self.driver_config).handle_action(action="put", name="test_save", journey=self.journey)

        journey_store_save_mock.assert_called_once_with(
            name="test_save", journey=self.journey
        )




"""
For ussd we need two store two things.
    1. users ussd session
    2. customer journey.


For ussd session each session has its own record and all users use the same journey.


users ussd session store
========================
comming soon


Customer journey store
======================

.. automodule:: ussd.store.journey_store

"""



"""
This test quit screen
"""
from ussd.tests import UssdTestCase


class TestQuitHandler(UssdTestCase.BaseUssdTestCase):
    validation_error_message = dict(
        example_of_quit_screen=dict(
            text=['This field is required.']
        )
    )

    def test(self):
        ussd_client = self.ussd_client()
        response = ussd_client.send('', raw=True)

        self.assertEqual(
            "Test getting variable from os environmen. variable_test",
            str(response)
        )

        self.assertFalse(response.status)



import attr


@attr.s
class Vertex(object):
    name = attr.ib()
    text = attr.ib(default="")
    id = attr.ib(default="")


@attr.s
class Link(object):
    start = attr.ib(factory=Vertex)
    end = attr.ib(factory=Vertex)
    text = attr.ib(factory=str)
    type = attr.ib(default="arrow")
    stroke = attr.ib(default="thick")


class Graph(object):

    def __init__(self):
        self.vertices = {}
        self.edges = []
        self.missing_vertices = {}

    def add_vertex(self, vertex: Vertex, **kwargs):
        self.vertices[vertex.name] = dict(id=vertex.name, text=vertex.text)
        self.vertices[vertex.name].update(kwargs)

        self.missing_vertices.pop(vertex.name, None)

    def add_link(self, link: Link):

        if link.start.name not in self.vertices:
            self.missing_vertices[link.start.name] = ''
        if link.end.name not in self.vertices:
            self.missing_vertices[link.end.name] = ''

        self.edges.append(
            dict(start=link.start.name, end=link.end.name,
                 text=link.text, type=link.type,
                 stroke=link.stroke)
        )

    def get_vertex(self, name):
        if isinstance(name, Vertex):
            name = name.name
        return self.vertices.get(name)

    def get_vertex_obj(self, name):
        raw_vertex = self.get_vertex(name)

        # add name
        raw_vertex['name'] = name

        return Vertex(**raw_vertex)

    def get_edges(self):
        return self.edges

    def convert_dict_to_link(self, raw_link: dict) -> Link:
        raw_link['start'] = self.get_vertex_obj(raw_link['start'])
        raw_link['end'] = self.get_vertex_obj(raw_link['end'])
        return Link(**raw_link)

    def __eq__(self, other):
        return self.vertices == other.vertices and \
               self.edges == other.edges


def get_mermaid_link_line(link: Link):
    if link.stroke == "dotted":
        link_line = "-.->" if link.text == "" \
            else '-."{link_text}".->'.format(link_text=link.text)
    else:
        link_line = "==>" if link.text == "" \
            else '=="{link_text}"==>'.format(link_text=link.text)
    return " " + link_line + " "


def get_mermaid_node_text(vertex: Vertex):
    text = vertex.id if vertex.text == "" else \
        '{id}["{text}"]'.format(id=vertex.id, text=vertex.text)

    # replace new line with <br>
    text = text.replace('\n', '<br>')

    return text


def add_mermaid_node_text(vertex: Vertex, added_texts: list):
    if vertex.name in added_texts:
        text = vertex.name
    else:
        text = get_mermaid_node_text(vertex)
        added_texts.append(vertex.name)
    return text


def convert_graph_to_mermaid_text(graph: Graph) -> str:
    added_texts = []

    mermaid_text = "graph TD\n"

    for i in graph.get_edges():
        link = graph.convert_dict_to_link(i)

        # add first node
        mermaid_text += add_mermaid_node_text(link.start, added_texts)

        # add link
        mermaid_text += get_mermaid_link_line(link)

        # add second node
        mermaid_text += add_mermaid_node_text(link.end, added_texts)

        # add end line
        mermaid_text += '\n'
    return mermaid_text



from ussd_version.version import VERSION
from ussd.tests import UssdTestCase


class TestScreenUsing(UssdTestCase.BaseUssdTestCase):
    validate_ussd = False

    def get_ussd_client(self):
        return self.ussd_client(
            generate_customer_journey=False,
            extra_payload={
                'journey_name': "sample_journey",
                'journey_version': "testing_using_built_in_functions"
            }
        )

    def test(self):
        client = self.get_ussd_client()
        # dial in
        response = client.send('1')

        self.assertEqual(
            "You are using this version v{version}".format(version=VERSION),
            response
        )



from ..journey_store import JourneyStore
from collections import OrderedDict

store = OrderedDict()


class DummyStore(JourneyStore):
    """
    This store should not be used in production only meant to test the interface
    """
    
    def __init__(self, user="default"):
        self.user = user
        if not store.get(self.user):
            store[user] = {}
        self.store = store[user]
            
    def _get(self, name, version, screen_name, **kwargs):
        if version == 'edit_mode':
            return self.store.get('edit_mode', {}).get(name)
        if self.store.get(name):
            if version is None:  # get the latest journey created
                journey = self.store[name][next(reversed(self.store[name]))]
            else:
                journey = self.store[name].get(version)
            if screen_name is not None:
                return journey.get(screen_name)
            return journey
        return None

    def _get_all_journey_version(self, name):
        return self.store.get(name, {})

    def _save(self, name, journey, version):
        if version == self.edit_mode_version:
            self.store['edit_mode'] = {name: journey}
        else:
            if self.store.get(name):
                self.store[name][version] = journey
            else:
                self.store[name] = OrderedDict({version: journey})
        return journey

    def _delete(self, name, version=None):
        if self.store.get(name):
            if self.store[name].get(version):
                del self.store[name][version]
            else:
                del self.store[name]
    
    def _all(self):
        return self.store
    
    def flush(self):
        pass




import typing
from marshmallow import fields


class UssdTextField(fields.Field):
    """
    Ussd text that's going to be displayed to the user.
    example

    text: This is the text that is going to be shown

    To display multi language

    text:
        en: This the text in english
        sw: this the text in swahili
        default: en
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if not isinstance(value, dict):
            return {
                'en': value,
                'default': 'en'}
        return value


class WithItemField(fields.Field):

    def _deserialize(
        self,
        value: typing.Any,
        attr: typing.Optional[str],
        data: typing.Optional[typing.Mapping[str, typing.Any]],
        **kwargs
    ):
        if isinstance(value, list):
            return value
        return value


class WithDictField(fields.Field):

    def _deserialize(
            self,
            value: typing.Any,
            attr: typing.Optional[str],
            data: typing.Optional[typing.Mapping[str, typing.Any]],
            **kwargs
    ):
        if isinstance(value, dict):
            return value
        return value


class NextUssdScreenField(fields.Field):

    def _deserialize(
        self,
        value: typing.Any,
        attr: typing.Optional[str],
        data: typing.Optional[typing.Mapping[str, typing.Any]],
        **kwargs
    ):
        if not isinstance(value, list):
            return [{"condition": "true", "next_screen": value}]
        return value



from ussd.core import UssdHandlerAbstract, UssdHandlerMetaClass
from ussd.utilities import str_to_class
import typing
from ussd.graph import Link, Vertex
from ussd.screens.schema import UssdBaseScreenSchema
from marshmallow import fields, INCLUDE, ValidationError, validates_schema


class ScreenObjField(fields.Field):

    def _deserialize(
        self,
        value: typing.Any,
        attr: typing.Optional[str],
        data: typing.Optional[typing.Mapping[str, typing.Any]],
        **kwargs
    ):
        try:
            screen_obj = str_to_class(value)
        except Exception as e:
            raise ValidationError(str(e))
        else:
            if not isinstance(screen_obj, UssdHandlerMetaClass):
                raise ValidationError(
                    "Screen object should be of type UssdHandlerAbstract"
                )
            return screen_obj


class CustomScreenSchema(UssdBaseScreenSchema):

    screen_obj = ScreenObjField(required=True)

    class Meta:
        unknown = INCLUDE

    @validates_schema
    def validate_custom_schema(self, data, **kwargs):
        screen_obj = data['screen_obj']
        if hasattr(screen_obj, 'serializer'):
            schema = screen_obj.serializer(context=self.context)
            schema.load(data=data)



class CustomScreen(UssdHandlerAbstract):
    """
    If you have a particular user case that's not yet covered by
    our existing screens, this is the screen to use.

    This screen allows us to define our own ussd screen.

    To create it you need the following fields.
        1. screen_object
            This is the path to be used to import the class
        2. serializer (optional)
            This if you want to be validating your screen with
            specific fields
        3. You can define any field that you feel
            your custom screen might need.
    EXAMPLE:
        examples of custom screen 
            .. code-block:: python
            
                class SampleCustomHandler1(UssdHandlerAbstract):
                    abstract = True  # don't register custom classes
                    @staticmethod
                    def show_ussd_content():  # This method doesn't have to be static
                        # Do anything custom here.
                        return UssdResponse("This is a custom Handler1")
                
                    def handle_ussd_input(self, ussd_input):
                        # Do anything custom here
                        print(ussd_input)  # pep 8 for the sake of using it.
                        return self.ussd_request.forward('custom_screen_2')
    
    
                class SampleSerializer(UssdBaseSerializer, NextUssdScreenSerializer):
                        input_identifier = serializers.CharField(max_length=100)
                
                
                class SampleCustomHandlerWithSerializer(UssdHandlerAbstract):
                    abstract = True  # don't register custom classes
                    serializer = SampleSerializer
                
                    @staticmethod
                    def show_ussd_content():  # This method doesn't have to be static
                        return "Enter a digit and it will be doubled on your behalf"
                
                    def handle_ussd_input(self, ussd_input):
                        self.ussd_request.session[
                            self.screen_content['input_identifier']
                        ] = int(ussd_input) * 2
                
                        return self.ussd_request.forward(
                            self.screen_content['next_screen']
                        )
        
        example of defining a yaml
            
            .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_menu_screen_conf.yml
    """
    screen_type = "custom_screen"
    serializer = CustomScreenSchema

    def __init__(self, *args, **kwargs):
        super(CustomScreen, self).__init__(*args, **kwargs)
        self.custom_screen_instance = str_to_class(
            self.screen_content['screen_obj']
        )(
            self.ussd_request,
            self.handler,
            self.screen_content,
            initial_screen={},
        )

    def handle(self):
        # Todo initiate the class in the core. to avoid double changing
        # if the parameter changes.

        # calling the custom screen handler method
        return self.custom_screen_instance.handle()

    def show_ussd_content(self, **kwargs):
        if self.custom_screen_instance.__class__.__dict__.get('show_ussd_content'):
            return self.custom_screen_instance.show_ussd_content()
        return "custom_screen\n{}".format(self.screen_content['screen_obj'])

    def get_next_screens(self) -> typing.List[Link]:
        if self.custom_screen_instance.__class__.__dict__.get('get_next_screens'):
            return self.custom_screen_instance.get_next_screens()
        return [
            Link(Vertex(self.handler),
                 Vertex(self.screen_content['next_screen']),
                 self.screen_content['input_identifier'])]



from ussd.core import UssdHandlerAbstract
import importlib
from ussd.graph import Link, Vertex
from ussd.screens.schema import NextUssdScreenSchema
from marshmallow import fields, ValidationError, INCLUDE
import typing


class FunctionField(fields.Field):

    def _deserialize(
        self,
        value: typing.Any,
        attr: typing.Optional[str],
        data: typing.Optional[typing.Mapping[str, typing.Any]],
        **kwargs
    ):
        split_path = value.split('.')
        if len(split_path) <= 1:
            raise ValidationError(
                "Module name where function is located not given"
            )
        function_name = split_path[-1]
        module_name = '.'.join(value.split('.')[:-1])
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            raise ValidationError(
                "Module {0} does not exist".format(module_name)
            )

        if not hasattr(module, function_name):
            raise ValidationError(
                "Function {0} does not exist".format(value)
            )
        return module


class FunctionScreenSerializer(NextUssdScreenSchema):
    """
    Fields used to create this screen:

    1. function
        This is the function that will be called at this screen.
    2. session_key
        Once your function has been called the output of your function
        will be saved in ussd session using session_key
    3. next_screen
        Once your function has been called this it goes to the
        screen specified in next_screen
    """
    session_key = fields.Str(required=True)
    function = FunctionField(required=True)

    class Meta:
        unknown = INCLUDE


class FunctionScreen(UssdHandlerAbstract):
    """
    This screen is invisible to the user. Its used to if you want to call a function you
    have implemented.

    Its like a complement of http screen. In http screen you make a request to an
    external service to perform some logic.

    This screen on the contrary if the logic that you want to be executed is within
    your application you use this screen to execute it.

    Your function will be called with UssdRequest object.
    And it should return a dictionary that will be saved in ussd session

    Below is the UssdRequest that will be used.
        .. autoclass:: ussd.core.UssdRequest

    Screen specification
        .. autoclass:: ussd.screens.function_screen.FunctionScreenSerializer

    Examples of function screens:
        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_function_screen_conf.yml
    """
    screen_type = "function_screen"
    serializer = FunctionScreenSerializer

    def handle(self):
        split_path = self.screen_content['function'].split('.')
        function_name = split_path[-1]
        module_name = '.'.join(split_path[:-1])
        module = importlib.import_module(module_name)

        self.ussd_request.session[
            self.screen_content['session_key']
        ] = getattr(module, function_name)(self.ussd_request)

        return self.route_options()

    def show_ussd_content(self, **kwargs):
        return "function_screen\n{}".format(self.screen_content['function'])

    def get_next_screens(self):
        links = []
        screen_vertex = Vertex(self.handler)
        if isinstance(self.screen_content.get("next_screen"), list):
            for i in self.screen_content.get("next_screen", []):
                links.append(
                    Link(screen_vertex,
                         Vertex(i['next_screen'], ""),
                         i['condition'])
            )
        elif self.screen_content.get('next_screen'):
            links.append(
                Link(
                    screen_vertex,
                    Vertex(self.screen_content['next_screen']),
                    self.screen_content['session_key']
                )
            )

        if self.screen_content.get('default_next_screen'):
            links.append(
                Link(
                    screen_vertex,
                    Vertex(self.screen_content['default_next_screen'], ""),
                    self.screen_content['session_key']
                )
            )
        return links



from ussd.core import UssdResponse
from ussd.utils.encoding import force_text
import re
from ussd.screens.menu_screen import MenuScreen
from ussd.graph import Link, Vertex
import typing
from ussd.screens.schema import UssdTextSchema, UssdContentBaseSchema, NextUssdScreenSchema, MenuOptionSchema, NextUssdScreenField
from marshmallow import fields


class InputValidatorSchema(UssdTextSchema):
    regex = fields.Str(required=False)
    expression = fields.Str(required=False)


class InputSchema(UssdContentBaseSchema, NextUssdScreenSchema):
    input_identifier = fields.Str(required=True)
    validators = fields.List(fields.Nested(InputValidatorSchema), required=False)
    options = fields.List(fields.Nested(MenuOptionSchema), required=False)
    default_next_screen = NextUssdScreenField(required=False)


class InputScreen(MenuScreen):
    """

    This screen prompts the user to enter an input

    Fields required:
        - text: this the text to display to the user.
        - input_identifier: input amount entered by users will be saved
                            with this key. To access this in the input
                            anywhere {{ input_identifier }}
        - next_screen: The next screen to go after the user enters
                        input
        - validators:
            - text: This is the message to display when the validation fails
              regex: regex used to validate ussd input. Its mutually exclusive
              with expression
            - expression: if regex is not enough you can use a jinja expression
             will be called ussd request object
              text: This the message thats going to be displayed if expression
              returns False
        - options (This field is optional):
            This is a list of options to display to the user
            each option is a key value pair of option text to display
            and next_screen to redirect if option is selected.
            Example of option:

            .. code-block:: yaml

                   options:
                    - text: option one
                      next_screen: screen_one
                    - text: option two
                      next_screen: screen_two

    Example:
        .. literalinclude:: .././ussd/tests/sample_screen_definition/valid_input_screen_conf.yml
    """

    screen_type = "input_screen"
    serializer = InputSchema

    def handle_invalid_input(self):
        # validate input
        validation_rules = self.screen_content.get("validators", {})
        for validation_rule in validation_rules:

            if 'regex' in validation_rule:
                regex_expression = validation_rule['regex']
                regex = re.compile(regex_expression)
                is_valid = bool(
                    regex.search(
                        force_text(self.ussd_request.input)
                    ))
            else:
                is_valid = self.evaluate_jija_expression(
                    validation_rule['expression'],
                    session=self.ussd_request.session
                )

            # show error message if validation failed
            if not is_valid:
                return UssdResponse(
                    self.get_text(
                        validation_rule['text']
                    )
                )

        self.ussd_request.session[
            self.screen_content['input_identifier']
        ] = self.ussd_request.input

        return self.route_options()

    def get_next_screens(self) -> typing.List[Link]:
        # generate validators links
        links = []
        screen_vertex = Vertex(self.handler)
        for index, validation_screen in enumerate(self.screen_content.get("validators", [])):
            validator_screen_name = self.handler + "_validator_" + str(index + 1)
            validation_vertex = Vertex(validator_screen_name,
                                       self.get_text(validation_screen['text']))
            if 'regex' in validation_screen:
                validation_command = 'regex: ' + validation_screen['regex']
            else:
                validation_command = 'expression: ' + validation_screen['expression']
            links.append(
                Link(screen_vertex,
                     validation_vertex,
                     "validation",
                     "arrow",
                     "dotted"
                     )
            )

            links.append(
                Link(
                    validation_vertex,
                    screen_vertex,
                    validation_command,
                    "arrow",
                    "dotted"
                )
            )

        if isinstance(self.screen_content.get("next_screen"), list):
            for i in self.screen_content.get("next_screen", []):
                links.append(
                    Link(screen_vertex,
                         Vertex(i['next_screen'], ""),
                         i['condition'])
            )
        elif self.screen_content.get('next_screen'):
            links.append(
                Link(
                    screen_vertex,
                    Vertex(self.screen_content['next_screen']),
                    self.screen_content['input_identifier']
                )
            )

        if self.screen_content.get('default_next_screen'):
            links.append(
                Link(
                    screen_vertex,
                    Vertex(self.screen_content['default_next_screen'], ""),
                    self.screen_content['input_identifier']
                )
            )
        return links + super(InputScreen, self).get_next_screens()



"""
This test initial screen
"""
from ussd.tests import UssdTestCase


class TestInitialHandler(UssdTestCase.BaseUssdTestCase):
    expected_error = FileNotFoundError
    validation_error_message = {}



from ..journey_store import JourneyStore
import boto3
from structlog import get_logger
from botocore.config import Config
from copy import deepcopy
from boto3.dynamodb.conditions import Key
import os

BOTO_CORE_CONFIG = os.environ.get('BOTO_CORE_CONFIG', None)
USE_LOCAL_DYNAMODB_SERVER = os.environ.get('USE_LOCAL_DYNAMODB_SERVER', False)

logger = get_logger(__name__)

# We'll find some better way to do this.
_DYNAMODB_CONN = None
_DYNAMODB_TABLE = {}

# defensive programming if config has been defined
# make sure it's the correct format.
if BOTO_CORE_CONFIG:
    assert isinstance(BOTO_CORE_CONFIG, Config)

dynamo_kwargs = dict(
    service_name='dynamodb',
    config=BOTO_CORE_CONFIG
)


def dynamodb_connection_factory(low_level=False, endpoint=None):
    """
    Since SessionStore is called for every single page view, we'd be
    establishing new connections so frequently that performance would be
    hugely impacted. We'll lazy-load this here on a per-worker basis. Since
    boto3.resource.('dynamodb')objects are state-less (aside from security
    tokens), we're not too concerned about thread safety issues.
    """
    boto_config = deepcopy(dynamo_kwargs)
    if endpoint is not None:
        boto_config["endpoint_url"] = endpoint

    if low_level:
        return boto3.client(**boto_config)

    global _DYNAMODB_CONN

    if not _DYNAMODB_CONN:
        logger.debug("Creating a DynamoDB connection.")
        _DYNAMODB_CONN = boto3.resource(**boto_config)
    return _DYNAMODB_CONN


def dynamodb_table(table: str, endpoint=None):
    global _DYNAMODB_TABLE

    if not _DYNAMODB_TABLE.get(table):
        _DYNAMODB_TABLE[table] = dynamodb_connection_factory(endpoint=endpoint).Table(table)
    return _DYNAMODB_TABLE[table]


def create_table(table_name: str):
    params = {
        'TableName': table_name,
        'KeySchema': [
            {'AttributeName': "username", 'KeyType': "HASH"},  # Partition key
            {'AttributeName': "journeyAndVersion", 'KeyType': "RANGE"}  # Sort key
        ],
        'AttributeDefinitions': [
            {'AttributeName': "username", 'AttributeType': "S"},
            {'AttributeName': "journeyAndVersion", 'AttributeType': "S"}
        ],
        'ProvisionedThroughput': {
            'ReadCapacityUnits': 10,
            'WriteCapacityUnits': 10
        }
    }

    dynamodb = dynamodb_connection_factory(low_level=True, endpoint="http://dynamodb:8000")
    # Create the table
    dynamodb.create_table(**params)

    # Wait for the table to exist before exiting
    print('Waiting for', table_name, '...')
    waiter = dynamodb.get_waiter('table_exists')
    waiter.wait(TableName=table_name)


def delete_table(table_name: str):
    connection = dynamodb_connection_factory(low_level=True, endpoint="http://dynamodb:8000")
    connection.delete_table(
        TableName=table_name
    )


def append_dynamo_key_with_user(func):
    def wrapper(*args, **kwargs):
        # we need to append the parameter name with user
        # first parameter is instance i.e self
        # second parameter is name
        if not args[1].startswith("{}#".format(args[0].user)):
            args = list(args)
            args[1] = "{0}#{1}".format(args[0].user, args[1])
            args = tuple(args)
        return func(*args, **kwargs)

    return wrapper


class DynamoDb(JourneyStore):
    edit_mode_version = "-1"
    hash_key = "username"
    sort_key = "journeyAndVersion"

    def __init__(self, table_name, endpoint=None, user="default"):
        self.table_name = table_name
        self.table = dynamodb_table(table_name, endpoint=endpoint)
        self.user = user

    @staticmethod
    def _sort_key_journey(name):
        return "{0}#".format(name)

    @staticmethod
    def _sort_key_version(version):
        return "{0}#".format(version)

    def _generate_sort_key(self, name, version):
        return "{0}{1}".format(self._sort_key_journey(name), self._sort_key_version(version))

    def _get(self, name, version, screen_name, **kwargs):
        screen_kwarg = {}
        if screen_name is not None:
            screen_kwarg["ProjectionExpression"] = f'journey.{screen_name}'

        if version is None:
            response = self.table.query(
                KeyConditionExpression=Key(self.hash_key).eq(self.user) &
                                       Key(self.sort_key).begins_with(self._sort_key_journey(name)),
                **kwargs
            )
            item = response.get("Items")[-1]
        else:
            key = {
                self.hash_key: self.user,
                self.sort_key: self._generate_sort_key(name, version)
            }

            response = self.table.get_item(Key=key, **screen_kwarg)
            item = response.get('Item')

        if item:
            item = item.get('journey', {})
            if screen_name:
                item = item.get(screen_name)
        return item or None

    def _get_all_journey_version(self, name):
        results = {}
        for i in self._query(name):
            version = i[self.sort_key].split("#")[1]
            results[version] = i.get('journey', {})
        return results

    def _query(self, name, **kwargs):
        response = self.table.query(
            KeyConditionExpression=Key(self.hash_key).eq(self.user) &
                                   Key(self.sort_key).begins_with(self._sort_key_journey(name)),
            **kwargs
        )
        return response.get('Items')

    def _save(self, name, journey, version):
        item = {
            self.hash_key: self.user,
            self.sort_key: self._generate_sort_key(name, version)
        }
        item.update({'journey': journey})

        response = self.table.put_item(
            Item=item
        )

    def _delete(self, name, version=None):
        items = [
            {
                self.hash_key: self.user,
                self.sort_key: self._generate_sort_key(name, version)
            }
        ]
        if version is None:
            items = self._query(name,
                                **{"ProjectionExpression": "{0}, {1}".format(self.hash_key, self.sort_key)})

        with self.table.batch_writer() as batch:
            for i in items:
                batch.delete_item(
                    Key={
                        self.hash_key: i[self.hash_key],
                        self.sort_key: i[self.sort_key]
                    }
                )

    def _all(self):
        response = self.table.query(
            KeyConditionExpression=Key(self.hash_key).eq(self.user)
        )

        results = dict()
        for i in response.get('Items'):
            name, version, others = i[self.sort_key].split("#")

            if not results.get(name):
                results[name] = {}

            results[name][version] = i.get('journey', {})

        return results

    def flush(self):
        all_records = self.table.scan()
        for i in all_records['Items']:
            self._delete(i[self.hash_key], i[self.sort_key])



"""
Comming soon
"""
import inspect
import json
import os
import re
import typing
from collections import namedtuple
from copy import copy
from datetime import datetime
from urllib.parse import unquote

import requests
from jinja2 import Environment
from structlog import get_logger

from ussd import defaults as ussd_airflow_variables
from ussd import utilities
from ussd.tasks import report_session
from .graph import Graph, Link, Vertex, convert_graph_to_mermaid_text
from ussd.screens.schema import UssdBaseScreenSchema
from ussd.store.journey_store import JourneyStore
from ussd.store.journey_store.YamlJourneyStore import YamlJourneyStore
from simplekv import KeyValueStore
from simplekv.fs import FilesystemStore
from ussd.session_store import SessionStore
from marshmallow.schema import SchemaMeta

_registered_ussd_handlers = {}
_registered_filters = {}
_customer_journey_files = []
_built_in_functions = {}

# initialize jinja2 environment
env = Environment(keep_trailing_newline=True)
env.filters.update(_registered_filters)


class MissingAttribute(Exception):
    pass


class InvalidAttribute(Exception):
    pass


class DuplicateSessionId(Exception):
    pass


def register_filter(func_name, *args, **kwargs):
    filter_name = func_name.__name__
    _registered_filters[filter_name] = func_name


def register_function(func_name, *args, **kwargs):
    function_name = func_name.__name__
    _built_in_functions[function_name] = func_name


class UssdRequest(object):
    """
    :param session_id:
        used to get session or create session if does not
        exits.

        If session is less than 8 we add *s* to make the session
        equal to 8

    :param phone_number:
        This the user identifier

    :param input:
        This ussd input the user has entered.

    :param language:
        Language to use to display ussd

    :param kwargs:
        Extra arguments.
        All the extra arguments will be set to the self attribute

        For instance:

        .. code-block:: python

            from ussd.core import UssdRequest

            ussdRequest = UssdRequest(
                '12345678', '702729654', '1', 'en',
                name='mwas'
            )

            # accessing kwarg argument
            ussdRequest.name
    """

    def __init__(self, session_id, phone_number,
                 ussd_input, language,
                 journey_name,
                 journey_store: JourneyStore = None,
                 journey_version=None,
                 session_store_backend: KeyValueStore = FilesystemStore("./session_data"),
                 default_language=None,
                 use_built_in_session_management=False,
                 expiry=180,
                 **kwargs):
        """
        :param session_id: Used to maintain session 
        :param phone_number: user dialing in   
        :param ussd_input: input entered by user
        :param language: language to be used
        :param default_language: language to used
        :param use_built_in_session_management: Used to enable ussd_airflow to 
            manage its own session, by default its set to False, is set to true 
        then the session_id should be None and expiry can't be None. 
        :param expiry: Its only used if use_built_in_session_management has
        been enabled. 
        :param kwargs: All other extra arguments
        """

        self.expiry = expiry
        # A bit of defensive programming to make sure
        # session_built_in_management has been initiated
        if use_built_in_session_management and session_id is not None:
            raise InvalidAttribute("When using built_in_session_management "
                                   "has been enabled session_id should "
                                   "be None")
        if use_built_in_session_management and expiry is None:
            raise InvalidAttribute("When built_in_session_management has been"
                                   "enabled expiry should not be None")
        # session id should not be None if built in session management
        # has not been enabled
        if session_id is None and not use_built_in_session_management:
            raise InvalidAttribute(
                "Session id should not be None if built in session management "
                "has not been enabled"
            )

        if session_id is None:
            session_id = str(phone_number)

        # for support when using django session table
        if len(str(session_id)) < 8:
            session_id = 's' * (8 - len(str(session_id))) + session_id

        self.phone_number = str(phone_number)
        self.input = unquote(ussd_input)
        self.language = language
        self.default_language = default_language or 'en'
        self.session_id = session_id

        # session store config
        self.use_built_in_session_management = use_built_in_session_management
        self.session_store_backend = session_store_backend
        self.session = self.get_session()
        self.session.set_expiry(self.expiry)

        # journey config
        if journey_store is None:
            self.journey_store = YamlJourneyStore("./ussd/tests/sample_screen_definition")
        else:
            self.journey_store = journey_store
        self.journey_name = journey_name
        self.journey_version = journey_version

        # screen configs
        self.menu_index_format = kwargs.get('menu_index_format', ". ")

        for key, value in kwargs.items():
            if not hasattr(self, key):
                setattr(self, key, value)

    def forward(self, handler_name):
        """
        Forwards a copy of the current request to a new
        handler. Clears any input, as it is assumed this was meant for
        the previous handler. If you need to pass info between
        handlers, do it through the USSD session.
        """
        new_request = copy(self)
        new_request.input = ''
        return new_request, handler_name

    def all_variables(self):
        all_variables = copy(self.__dict__)

        # delete session if it exist
        all_variables.pop("session", None)

        return all_variables

    def built_in_session_management(self):
        if self.session_id is None:
            raise TypeError("Session id should not be null")

        session = self.get_session_from_store()

        # get last time session was updated
        if session.get_expiry_age() <= 0:
            # clear this session id
            # and save the data in another session key
            previous_session_key = session.cycle_data()
            session[ussd_airflow_variables.previous_session_id] = previous_session_key
        return session

    def get_session_from_store(self) -> SessionStore:
        return SessionStore(session_key=self.session_id,
                            kv_store=self.session_store_backend)

    def get_session(self) -> SessionStore:
        if self.use_built_in_session_management:
            return self.built_in_session_management()
        return self.get_session_from_store()

    def get_screens(self, screen_name=None):
        return self.journey_store.get(
            self.journey_name,
            self.journey_version,
            screen_name
        )


class UssdResponse(object):
    """
    :param text:
        This is the ussd text to display to the user
    :param status:
        This shows the status of ussd session.

        True -> to continue with the session

        False -> to end the session
    :param session:
        This is the session object of the ussd session
    """

    def __init__(self, text, status=True, session=None):
        self.text = text
        self.status = status
        self.session = session

    def dumps(self):
        return self.text

    def __str__(self):
        return self.dumps()


class UssdHandlerMetaClass(type):

    def __init__(cls, name, bases, attr, **kwargs):
        super(UssdHandlerMetaClass, cls).__init__(
            name, bases, attr)

        abstract = attr.get('abstract', False)

        if not abstract or attr.get('screen_type', '') == 'custom_screen':
            required_attributes = ('screen_type', 'serializer', 'handle')

            # check all attributes have been defined
            for attribute in required_attributes:
                if attribute not in attr and not hasattr(cls, attribute):
                    raise MissingAttribute(
                        "{0} is required in class {1}".format(
                            attribute, name)
                    )

            if not isinstance(attr['serializer'], SchemaMeta):
                raise InvalidAttribute(
                    "serializer should be a "
                    "instance of {serializer}".format(
                        serializer=SchemaMeta)
                )
            _registered_ussd_handlers[attr['screen_type']] = cls


class UssdHandlerAbstract(object, metaclass=UssdHandlerMetaClass):
    abstract = True

    def __init__(self, ussd_request: UssdRequest,
                 handler: str, screen_content: dict,
                 initial_screen: dict, logger=None,
                 raw_text=False):
        self.ussd_request = ussd_request
        self.handler = handler
        self.screen_content = screen_content
        self.raw_text = raw_text

        self.SINGLE_VAR = re.compile(r"^%s\s*(\w*)\s*%s$" % (
            '{{', '}}'))
        self.clean_regex = re.compile(r'^{{\s*(\S*)\s*}}$')
        self.logger = logger or get_logger(__name__).bind(
            handler=self.handler,
            screen_type=getattr(self, 'screen_type', 'custom_screen'),
            **self.ussd_request.all_variables(),
        )
        self.initial_screen = initial_screen

        self.pagination_config = self.initial_screen.get('pagination_config',
                                                         {})

        self.pagination_more_option = self._add_end_line(
            self.get_text(
                self.pagination_config.get('more_option', "more\n")
            )
        )
        self.pagination_back_option = self._add_end_line(
            self.get_text(
                self.pagination_config.get('back_option', "back\n")
            )
        )
        self.ussd_text_limit = self.pagination_config. \
            get("ussd_text_limit", ussd_airflow_variables.ussd_text_limit)

    def handle(self):
        if not self.ussd_request.input:
            ussd_response = self.show_ussd_content()
            return ussd_response if isinstance(ussd_response, UssdResponse) \
                else UssdResponse(str(ussd_response))
        return self.handle_ussd_input(self.ussd_request.input)

    def get_text_limit(self):
        return self.ussd_text_limit

    def show_ussd_content(self, **kwargs):
        raise NotImplementedError

    def handle_ussd_input(self, ussd_input):
        raise NotImplementedError

    def route_options(self, route_options=None):
        """
        iterates all the options executing expression comand.
        """
        if route_options is None:
            route_options = self.screen_content["next_screen"]

        if isinstance(route_options, str):
            return self.ussd_request.forward(route_options)

        loop_items = [0]
        if self.screen_content.get("with_items"):
            loop_items = self.evaluate_jija_expression(
                self.screen_content["with_items"],
                session=self.ussd_request.session
            ) or loop_items

        for item in loop_items:
            extra_context = {
                "item": item
            }
            if isinstance(loop_items, dict):
                extra_context.update(
                    dict(
                        key=item,
                        value=loop_items[item],
                        item={item: loop_items[item]}
                    )
                )

            for option in route_options:
                if self.evaluate_jija_expression(
                        option.get('expression') or option['condition'],
                        session=self.ussd_request.session,
                        extra_context=extra_context
                ):
                    return self.ussd_request.forward(option['next_screen'])
        return self.ussd_request.forward(
            self.screen_content['default_next_screen']
        )

    @staticmethod
    def get_session_items(session) -> dict:
        return dict(iter(session.items()))

    @classmethod
    def get_context(cls, session, extra_context=None):
        context = cls.get_session_items(session)

        context.update(
            dict(os.environ)
        )

        if extra_context is not None:
            context.update(extra_context)

        # add timestamp in the context
        context.update(
            dict(now=datetime.now())
        )

        # add all built in functions
        context.update(
            _built_in_functions
        )
        return context

    @staticmethod
    def render_text(session, text, context=None, extra=None, encode=None):
        if context is None:
            context = UssdHandlerAbstract.get_context(
                session
            )

        if extra:
            context.update(extra)

        template = env.from_string(text or '')
        text = template.render(context)
        return json.dumps(text) if encode is 'json' else text

    def get_text(self, text_context=None):
        text_context = self.screen_content.get('text') \
            if text_context is None \
            else text_context

        if isinstance(text_context, dict):
            language = (self.ussd_request.session.get('override_language') or self.ussd_request.language) \
                if self.ussd_request.language \
                   in text_context.keys() \
                else self.ussd_request.default_language
            text_context = text_context[language]

        if self.raw_text:
            return text_context
        return self.render_text(
            self.ussd_request.session,
            text_context
        )

    @classmethod
    def evaluate_jija_expression(cls, expression, session,
                                 extra_context=None,
                                 lazy_evaluating=False,
                                 default=None):
        if not isinstance(expression, str) or \
                (lazy_evaluating and not cls._contains_vars(
                    expression)):
            return expression

        context = cls.get_context(
            session, extra_context=extra_context)

        try:
            expr = env.compile_expression(
                expression.replace("{{", "").replace("}}", "")
            )
            return expr(context)
        except Exception:
            try:
                return env.from_string(expression or '').render(context)
            except Exception:
                return default

    @classmethod
    def validate(cls, screen_name: str, ussd_content: dict) -> (bool, dict):
        screen_content = ussd_content[screen_name]
        schema = cls.serializer(context=ussd_content)
        errors = schema.validate(screen_content)
        return False if errors else True, errors

    @staticmethod
    def _contains_vars(data):
        '''
        returns True if the data contains a variable pattern
        '''
        if isinstance(data, str):
            for marker in ('{%', '{{', '{#'):
                if marker in data:
                    return True
        return False

    @staticmethod
    def _add_end_line(text):
        if text and '\n' not in text:
            text += '\n'
        return text

    def get_loop_items(self):
        loop_items = self.evaluate_jija_expression(
            self.screen_content["with_items"],
            session=self.ussd_request.session
        ) if self.screen_content.get("with_items") else [0] or [0]
        return loop_items

    @classmethod
    def render_request_conf(cls, session, data):
        if isinstance(data, str):
            jinja_results = cls.evaluate_jija_expression(data, session)
            return data if jinja_results is None else jinja_results

        elif isinstance(data, list):
            list_data = []
            for i in data:
                list_data.append(cls.render_request_conf(
                    session, i))

            return list_data

        elif isinstance(data, dict):
            dict_data = {}
            for key, value in data.items():
                dict_data.update(
                    {key: cls.render_request_conf(
                        session, value)}
                )
            return dict_data
        else:
            return data

    @staticmethod
    def get_variables_from_response_obj(response):
        response_varialbes = {}

        for i in inspect.getmembers(response):
            # Ignores anything starting with underscore
            # (that is, private and protected attributes)
            if not i[0].startswith('_'):
                # Ignores methods
                if not inspect.ismethod(i[1]) and \
                        type(i[1]) in \
                        (str, dict, int, dict, float, list, tuple):
                    if len(i) == 2:
                        response_varialbes.update(
                            {i[0]: i[1]}
                        )

        try:
            response_content = json.loads(response.content.decode())
        except json.JSONDecodeError:
            response_content = response.content.decode()

        if isinstance(response_content, dict):
            response_varialbes.update(
                response_content
            )

        # update content to save the one that has been decoded
        response_varialbes.update(
            {"content": response_content}
        )

        return response_varialbes

    @classmethod
    def make_request(cls, http_request_conf, response_session_key_save,
                     session, logger=None
                     ):
        logger = logger or get_logger(__name__).bind(
            action="make_request",
            session_id=session.session_key
        )
        logger.info("sending_request", **http_request_conf)
        response = requests.request(**http_request_conf)
        logger.info("response", status_code=response.status_code,
                    content=response.content)

        response_to_save = cls.get_variables_from_response_obj(response)

        # save response in session
        session[response_session_key_save] = response_to_save

        return response

    @staticmethod
    def fire_ussd_report_session_task(initial_screen: dict, session_id: str,
                                      support_countdown=True):
        ussd_report_session = initial_screen['ussd_report_session']
        args = (session_id,)
        kwargs = {'screen_content': initial_screen}
        keyword_args = ussd_report_session.get("async_parameters",
                                               {"countdown": 900}).copy()
        if not support_countdown and keyword_args.get('countdown'):
            del keyword_args['countdown']

        report_session.apply_async(
            args=args,
            kwargs=kwargs,
            **keyword_args
        )

    @staticmethod
    def get_handler(screen_type):
        return _registered_ussd_handlers[screen_type]

    def get_next_screens(self) -> typing.List[Link]:
        raise NotImplementedError

    def render_graph(self, ussd_journey: dict, graph: Graph):
        # adding the screen as vertex
        graph.add_vertex(Vertex(self.handler, self.show_ussd_content()))

        # get next screens
        next_screens = self.get_next_screens()

        # add links
        [graph.add_link(i) for i in next_screens]

        for i in next_screens:

            # check if node has been created
            if not (graph.get_vertex(i.end) and graph.get_vertex(i.end)['text'] != ""):
                graph.add_vertex(i.end)
                if ussd_journey.get(i.end.name):
                    next_screen_content = ussd_journey[i.end.name]
                    handler = self.get_handler(next_screen_content['type'])
                    handler(self.ussd_request, i.end.name, next_screen_content,
                            self.initial_screen, raw_text=self.raw_text). \
                        render_graph(ussd_journey, graph)


NextScreens = namedtuple("NextScreens", "next_screens links")


class UssdEngine(object):

    def __init__(self, ussd_request: UssdRequest):
        self.ussd_request = ussd_request
        initial_screen = ussd_request.get_screens('initial_screen')
        self.initial_screen = initial_screen \
            if isinstance(initial_screen, dict) \
            else {"initial_screen": initial_screen}
        self.logger = get_logger(__name__).bind(**ussd_request.all_variables())

    def ussd_dispatcher(self):

        # Clear input and initialize session if we are starting up
        if '_ussd_state' not in self.ussd_request.session:
            self.ussd_request.input = ''
            self.ussd_request.session['_ussd_state'] = {'next_screen': ''}
            self.ussd_request.session['ussd_interaction'] = []
            self.ussd_request.session['posted'] = False
            self.ussd_request.session['submit_data'] = {}
            self.ussd_request.session['session_id'] = self.ussd_request.session_id
            self.ussd_request.session['phone_number'] = self.ussd_request.phone_number

        # update self.ussd_request variable to session and template variables
        # to be used later for jinja2 evaluation
        self.ussd_request.session.update(self.ussd_request.all_variables())

        # for backward compatibility
        # there are some jinja template using ussd_request
        # eg. {{ussd_request.session_id}}
        self.ussd_request.session.update(
            {"ussd_request": self.ussd_request.all_variables()}
        )

        self.logger.debug('gateway_request', text=self.ussd_request.input)

        # Invoke handlers
        ussd_response = self.run_handlers()

        # Save session
        self.ussd_request.session.save()
        self.logger.debug('gateway_response', text=ussd_response.dumps(),
                          input="{redacted}")

        return ussd_response

    def run_handlers(self):

        handler = self.ussd_request.session['_ussd_state']['next_screen'] \
            if self.ussd_request.session.get('_ussd_state', {}).get('next_screen') \
            else "initial_screen"

        ussd_response = (self.ussd_request, handler)

        if handler != "initial_screen":
            # get start time
            start_time = utilities.string_to_datetime(
                self.ussd_request.session["ussd_interaction"][-1]["start_time"])
            end_time = datetime.now()
            # Report in milliseconds
            duration = (end_time - start_time).total_seconds() * 1000
            self.ussd_request.session["ussd_interaction"][-1].update(
                {
                    "input": self.ussd_request.input,
                    "end_time": utilities.datetime_to_string(end_time),
                    "duration": duration
                }
            )

        # Handle any forwarded Requests; loop until a Response is
        # eventually returned.
        while not isinstance(ussd_response, UssdResponse):
            self.ussd_request, handler = ussd_response

            screen_content = self.ussd_request.get_screens(handler)

            screen_type = 'initial_screen' \
                if handler == "initial_screen" and \
                   isinstance(screen_content, str) \
                else screen_content['type']

            ussd_response = _registered_ussd_handlers[screen_type](
                self.ussd_request,
                handler,
                screen_content,
                initial_screen=self.initial_screen,
                logger=self.logger
            ).handle()

        self.ussd_request.session['_ussd_state']['next_screen'] = handler

        self.ussd_request.session['ussd_interaction'].append(
            {
                "screen_name": handler,
                "screen_text": str(ussd_response),
                "input": self.ussd_request.input,
                "start_time": utilities.datetime_to_string(datetime.now())
            }
        )
        # Attach session to outgoing response
        ussd_response.session = self.ussd_request.session

        return ussd_response

    @staticmethod
    def validate_ussd_journey(ussd_content: dict) -> (bool, dict):
        errors = {}
        is_valid = True

        if not isinstance(ussd_content, dict):
            return False, {"_schema": ["journey should be a dictionary"]}

        # should define initial screen
        if not 'initial_screen' in ussd_content.keys():
            is_valid = False
            errors.update(
                {'hidden_fields': {
                    "initial_screen": ["This field is required."]
                }}
            )
        for screen_name, screen_content in ussd_content.items():
            # all screens should have type attribute
            if screen_name == "initial_screen" and \
                    isinstance(screen_content, str):
                if not screen_content in ussd_content.keys():
                    is_valid = False
                    errors.update(
                        dict(
                            screen_name="Screen not available"
                        )
                    )
                continue

            screen_type = screen_content.get('type')

            # all screen should have type field.
            base_schema = UssdBaseScreenSchema(context=ussd_content)
            base_errors = base_schema.validate(dict(type=screen_type))

            if base_errors:
                errors.update(
                    {screen_name: base_errors}
                )
                is_valid = False
                continue

            # all screen type have their handlers
            handlers = _registered_ussd_handlers[screen_type]

            screen_validation, screen_errors = handlers.validate(
                screen_name,
                ussd_content
            )
            if screen_errors:
                errors.update(
                    {screen_name: screen_errors}
                )

            if not screen_validation:
                is_valid = screen_validation

        return is_valid, errors

    @staticmethod
    def get_initial_screen(ussd_content: dict):
        initial_screen = ussd_content['initial_screen']
        return initial_screen if isinstance(initial_screen, dict) \
            else {"initial_screen": initial_screen}


def convert_error_response_to_mermaid_error(error_response: dict, errors=None, paths=None) -> list:
    errors = [] if errors is None else errors
    paths = [] if paths is None else paths

    for key, value in error_response.items():
        if isinstance(value, list):
            if len(value) == 1 and value[0] == "This field is required.":
                errors.append(
                    dict(path=paths, message=value[0].replace('This field', key))
                )
            else:
                errors.append(
                    dict(path=paths + [key], message='\n'.join(value))
                )
        else:
            convert_error_response_to_mermaid_error(value, errors, paths=paths + [key])
    return errors


def _convert_obj_to_mermaid_error_obj(key, value, errors: list):
    if isinstance(value, list):
        return dict(path=[key], message='\n'.join(value))

    for k, v in value.items():
        error_obj = {}
        j = _convert_obj_to_mermaid_error_obj(k, v)
        error_obj['path'] = [key] + j['path']
        error_obj['message'] = j['message']
        errors.append(error_obj)

    return error_obj


def render_journey_as_graph(ussd_screen: dict) -> Graph:
    graph = Graph()

    initial_screen = ussd_screen['initial_screen']
    _registered_ussd_handlers['initial_screen'](
        UssdRequest("dummy", "dummy", "", "en",
                    journey_name="", journey_version=""),
        "initial_screen",
        initial_screen,
        initial_screen,
        raw_text=True
    ).render_graph(ussd_screen, graph)
    return graph


def render_journey_as_mermaid_text(ussd_screen: dict) -> str:
    return convert_graph_to_mermaid_text(
        render_journey_as_graph(ussd_screen)
    )



"""
For session store we will rely on
https://pythonhosted.org/simplekv/#
"""
from simplekv import KeyValueStore
from simplekv.fs import FilesystemStore
import random
import string
import base64
from datetime import datetime, timedelta
import json
from collections import OrderedDict
from ussd import defaults as ussd_airflow_variables

def get_random_string(length=12,
                      allowed_chars='abcdefghijklmnopqrstuvwxyz'
                                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    """
    Return a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    return ''.join(random.choice(allowed_chars) for i in range(length))


date_format = '%Y-%m-%dT%H:%M:%S.%f'


class CustomJsonEncoder(json.JSONEncoder):

    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            if isinstance(obj, datetime):
                return obj.strftime(date_format)
            return obj.__class__.__name__


def datetime_parser(dct):
    dct = OrderedDict(dct)
    if dct.get(ussd_airflow_variables.session_expiry):
        dct[ussd_airflow_variables.session_expiry] = \
            datetime.strptime(dct[ussd_airflow_variables.session_expiry], date_format)
    return dct


class JSONSerializer:
    """
    Simple wrapper around json to be used in signing.dumps and
    signing.loads.
    """
    def dumps(self, obj):
        return json.dumps(obj, separators=(',', ':'), cls=CustomJsonEncoder).encode('latin-1')

    def loads(self, data):
        return json.loads(data.decode('latin-1'), object_pairs_hook=datetime_parser)


# session_key should not be case sensitive because some backends can store it
# on case insensitive file systems.
VALID_KEY_CHARS = string.ascii_lowercase + string.digits


SESSION_COOKIE_AGE = 180


class CreateError(Exception):
    """
    Used internally as a consistent exception type to catch from save (see the
    docstring for SessionBase.save() for details).
    """
    pass


class UpdateError(Exception):
    """
    Occurs if ussd tries to update a session that was deleted.
    """


class SessionStore(object):

    __not_given = object()

    def __init__(self, session_key=None,
                 kv_store: KeyValueStore = None,
                 default_session_cookie_age: int = SESSION_COOKIE_AGE):
        if kv_store is None:
            kv_store = FilesystemStore("./.session_data")
        self._session_key = session_key
        self.accessed = False
        self.modified = False
        self.serializer = JSONSerializer
        self.defaul_session_cookie_age = default_session_cookie_age
        self.kv_store = kv_store

    def __contains__(self, key):
        return key in self._session

    def __getitem__(self, key):
        return self._session[key]

    def __setitem__(self, key, value):
        self._session[key] = value
        self.modified = True

    def __delitem__(self, key):
        del self._session[key]
        self.modified = True

    def load(self):
        try:
            data = self.kv_store.get(self.session_key)
            return self.decode(data)
        except KeyError:
            return {}

    def exists(self, session_key):
        try:
            self.kv_store.get(session_key)
            return True
        except KeyError:
            return False

    def get(self, key, default=None):
        return self._session.get(key, default)

    def pop(self, key, default=__not_given):
        self.modified = self.modified or key in self._session
        args = () if default is self.__not_given else (default,)
        return self._session.pop(key, *args)

    def setdefault(self, key, value):
        if key in self._session:
            return self._session[key]
        else:
            self.modified = True
            self._session[key] = value
            return value

    def delete(self, session_key=None):
        if session_key is None:
            if self.session_key is None:
                return
            session_key = self.session_key
        self.kv_store.delete(session_key)

    def create(self):
        self._session_key = self._get_new_session_key()
        self.save(must_create=True)

    def save(self, must_create=False):
        if self.session_key is None:
            return self.create()
        data = self._get_session(no_load=must_create)

        self.set_expiry(self.get_expiry_date())
        self.kv_store.put(self.session_key, self.encode(data))

    def update(self, dict_):
        self._session.update(dict_)
        self.modified = True

    def has_key(self, key):
        return key in self._session

    def keys(self):
        return self._session.keys()

    def values(self):
        return self._session.values()

    def items(self):
        return self._session.items()

    def key_pair(self):
        return self._session

    def clear(self):
        # To avoid unnecessary persistent storage accesses, we set up the
        # internals directly (loading data wastes time, since we are going to
        # set it to an empty dict anyway).
        self._session_cache = {}
        self.accessed = True
        self.modified = True

    def encode(self, session_dict):
        """Return the given session dictionary serialized and encoded as a string."""
        serialized = self.serializer().dumps(session_dict)
        return base64.b64encode(serialized)

    def decode(self, session_data):
        encoded_data = base64.b64decode(session_data)
        return self.serializer().loads(encoded_data)

    def _get_new_session_key(self):
        "Return session key that isn't being used."
        while True:
            session_key = get_random_string(32, VALID_KEY_CHARS)
            if not self.exists(session_key):
                return session_key

    def _get_session(self, no_load=False):
        """
        Lazily load session from storage (unless "no_load" is True, when only
        an empty dict is stored) and store it in the current instance.
        """
        self.accessed = True
        try:
            return self._session_cache
        except AttributeError:
            if self.session_key is None or no_load:
                self._session_cache = {}
            else:
                self._session_cache = self.load()
        return self._session_cache

    @staticmethod
    def _validate_session_key(key):
        """
        Key must be truthy and at least 8 characters long. 8 characters is an
        arbitrary lower bound for some minimal key security.
        """
        return key and len(key) >= 8

    def _set_session_key(self, value):
        """
        Validate session key on assignment. Invalid values will set to None.
        """
        if self._validate_session_key(value):
            self.__session_key = value
        else:
            self.__session_key = None

    def _get_session_key(self):
        return self.__session_key

    _session_key = property(_get_session_key, _set_session_key)
    session_key = property(_get_session_key)
    _session = property(_get_session)

    def get_expiry_age(self, **kwargs):
        """Get the number of seconds until the session expires.

        Optionally, this function accepts `modification` and `expiry` keyword
        arguments specifying the modification and expiry of the session.
        """
        try:
            modification = kwargs['modification']
        except KeyError:
            modification = datetime.now()
        # Make the difference between "expiry=None passed in kwargs" and
        # "expiry not passed in kwargs", in order to guarantee not to trigger
        # self.load() when expiry is provided.
        try:
            expiry = kwargs['expiry']
        except KeyError:
            expiry = self.get('_session_expiry')

        if not expiry:   # Checks both None and 0 cases
            return self.defaul_session_cookie_age
        if not isinstance(expiry, datetime):
            return expiry
        delta = expiry - modification
        return delta.days * 86400 + delta.seconds

    def get_expiry_date(self, **kwargs):
        """Get session the expiry date (as a datetime object).

        Optionally, this function accepts `modification` and `expiry` keyword
        arguments specifying the modification and expiry of the session.
        """
        try:
            modification = kwargs['modification']
        except KeyError:
            modification = datetime.now()
        # Same comment as in get_expiry_age
        try:
            expiry = kwargs['expiry']
        except KeyError:
            expiry = self.get(ussd_airflow_variables.session_expiry)

        if isinstance(expiry, datetime):
            return expiry
        expiry = expiry or SESSION_COOKIE_AGE   # Checks both None and 0 cases
        return modification + timedelta(seconds=expiry)

    def set_expiry(self, value):
        """
        Set a custom expiration for the session. ``value`` can be an integer,
        a Python ``datetime`` or ``timedelta`` object or ``None``.

        If ``value`` is an integer, the session will expire after that many
        seconds of inactivity. If set to ``0`` then the session will expire on
        browser close.

        If ``value`` is a ``datetime`` or ``timedelta`` object, the session
        will expire at that specific future time.

        If ``value`` is ``None``, the session uses the global session expiry
        policy.
        """
        if value is None:
            # Remove any custom expiration for this session.
            try:
                del self['_session_expiry']
            except KeyError:
                pass
            return
        if isinstance(value, timedelta):
            value = datetime.now() + value
        self[ussd_airflow_variables.session_expiry] = value

    def flush(self):
        """
        Remove the current session data from the database and regenerate the
        key.
        """
        self.clear()
        self.delete()
        self._session_key = None

    def cycle_key(self):
        """
        Create a new session key, while retaining the current session data.
        """
        data = self._session
        key = self.session_key
        self.create()
        self._session_cache = data
        if key:
            self.delete(key)

    def cycle_data(self):
        """
        Creates a new session key saves the current data to that session key clears data of
        the current session key
        :return:
        """
        original_key = self.session_key
        # will set a new key and save data to that key
        self.create()
        new_key = self.session_key

        self._session_key = original_key
        self.clear()
        self.save()

        return new_key



from unittest import TestCase
from simplekv.fs import FilesystemStore
from ussd.session_store import SessionStore, SESSION_COOKIE_AGE
from datetime import datetime, timedelta
from freezegun import freeze_time
import uuid


class SessionTest(object):
    class SessionTestsMixin(TestCase):

        def backend(self, session_key=None) -> SessionStore:
            raise NotImplemented

        def setUp(self):
            self.session = self.backend()

        def tearDown(self):
            # NB: be careful to delete any sessions created; stale sessions fill up
            # the /tmp (with some backends) and eventually overwhelm it after lots
            # of runs (think buildbots)
            self.session.delete()

        def test_new_session(self):
            self.assertIs(self.session.modified, False)
            self.assertIs(self.session.accessed, False)

        def test_get_empty(self):
            self.assertIsNone(self.session.get('cat'))

        def test_store(self):
            self.session['cat'] = "dog"
            self.assertIs(self.session.modified, True)
            self.assertEqual(self.session.pop('cat'), 'dog')

        def test_pop(self):
            self.session['some key'] = 'exists'
            # Need to reset these to pretend we haven't accessed it:
            self.accessed = False
            self.modified = False

            self.assertEqual(self.session.pop('some key'), 'exists')
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, True)
            self.assertIsNone(self.session.get('some key'))

        def test_pop_default(self):
            self.assertEqual(self.session.pop('some key', 'does not exist'),
                             'does not exist')
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_pop_default_named_argument(self):
            self.assertEqual(self.session.pop('some key', default='does not exist'), 'does not exist')
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_pop_no_default_keyerror_raised(self):
            with self.assertRaises(KeyError):
                self.session.pop('some key')

        def test_setdefault(self):
            self.assertEqual(self.session.setdefault('foo', 'bar'), 'bar')
            self.assertEqual(self.session.setdefault('foo', 'baz'), 'bar')
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, True)

        def test_update(self):
            self.session.update({'update key': 1})
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, True)
            self.assertEqual(self.session.get('update key', None), 1)

        def test_has_key(self):
            self.session['some key'] = 1
            self.session.modified = False
            self.session.accessed = False
            self.assertIn('some key', self.session)
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_values(self):
            self.assertEqual(list(self.session.values()), [])
            self.assertIs(self.session.accessed, True)
            self.session['some key'] = 1
            self.session.modified = False
            self.session.accessed = False
            self.assertEqual(list(self.session.values()), [1])
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_keys(self):
            self.session['x'] = 1
            self.session.modified = False
            self.session.accessed = False
            self.assertEqual(list(self.session.keys()), ['x'])
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_items(self):
            self.session['x'] = 1
            self.session.modified = False
            self.session.accessed = False
            self.assertEqual(list(self.session.items()), [('x', 1)])
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, False)

        def test_clear(self):
            self.session['x'] = 1
            self.session.modified = False
            self.session.accessed = False
            self.assertEqual(list(self.session.items()), [('x', 1)])
            self.session.clear()
            self.assertEqual(list(self.session.items()), [])
            self.assertIs(self.session.accessed, True)
            self.assertIs(self.session.modified, True)

        def test_save(self):
            self.session.save()
            self.assertIs(self.session.exists(self.session.session_key), True)

        def test_delete(self):
            self.session.save()
            self.session.delete(self.session.session_key)
            self.assertIs(self.session.exists(self.session.session_key), False)

        def test_flush(self):
            self.session['foo'] = 'bar'
            self.session.save()
            prev_key = self.session.session_key
            self.session.flush()
            self.assertIs(self.session.exists(prev_key), False)
            self.assertNotEqual(self.session.session_key, prev_key)
            self.assertIsNone(self.session.session_key)
            self.assertIs(self.session.modified, True)
            self.assertIs(self.session.accessed, True)

        def test_cycle(self):
            self.session['a'], self.session['b'] = 'c', 'd'
            self.session.save()
            prev_key = self.session.session_key
            prev_data = list(self.session.items())
            self.session.cycle_key()
            self.assertIs(self.session.exists(prev_key), False)
            self.assertNotEqual(self.session.session_key, prev_key)
            self.assertEqual(list(self.session.items()), prev_data)

        def test_cycle_with_no_session_cache(self):
            self.session['a'], self.session['b'] = 'c', 'd'
            self.session.save()
            prev_data = self.session.items()
            self.session = self.backend(self.session.session_key)
            self.assertIs(hasattr(self.session, '_session_cache'), False)
            self.session.cycle_key()
            self.assertCountEqual(self.session.items(), prev_data)

        def test_save_doesnt_clear_data(self):
            self.session['a'] = 'b'
            self.session.save()
            self.assertEqual(self.session['a'], 'b')

        def test_invalid_key(self):
            # Submitting an invalid session key (either by guessing, or if the db has
            # removed the key) results in a new key being generated.
            try:
                session = self.backend('1')
                session.save()
                self.assertNotEqual(session.session_key, '1')
                self.assertIsNone(session.get('cat'))
                session.delete()
            finally:
                # Some backends leave a stale cache entry for the invalid
                # session key; make sure that entry is manually deleted
                session.delete('1')

        def test_session_key_empty_string_invalid(self):
            """Falsey values (Such as an empty string) are rejected."""
            self.session._session_key = ''
            self.assertIsNone(self.session.session_key)

        def test_session_key_too_short_invalid(self):
            """Strings shorter than 8 characters are rejected."""
            self.session._session_key = '1234567'
            self.assertIsNone(self.session.session_key)

        def test_session_key_valid_string_saved(self):
            """Strings of length 8 and up are accepted and stored."""
            self.session._session_key = '12345678'
            self.assertEqual(self.session.session_key, '12345678')

        def test_session_key_is_read_only(self):
            def set_session_key(session):
                session.session_key = session._get_new_session_key()
            with self.assertRaises(AttributeError):
                set_session_key(self.session)

        # Custom session expiry
        def test_default_expiry(self):
            # A normal session has a max age equal to settings
            self.assertEqual(self.session.get_expiry_age(), SESSION_COOKIE_AGE)

            # So does a custom session with an idle expiration time of 0 (but it'll
            # expire at browser close)
            self.session.set_expiry(0)
            self.assertEqual(self.session.get_expiry_age(), SESSION_COOKIE_AGE)

        def test_custom_expiry_seconds(self):
            modification = datetime.now()

            self.session.set_expiry(10)

            date = self.session.get_expiry_date(modification=modification)
            self.assertEqual(date, modification + timedelta(seconds=10))

            age = self.session.get_expiry_age(modification=modification)
            self.assertEqual(age, 10)

        def test_custom_expiry_timedelta(self):
            modification = datetime.now()
            with freeze_time(modification):
                self.session.set_expiry(timedelta(seconds=10))

                date = self.session.get_expiry_date(modification=modification)
                self.assertEqual(date, modification + timedelta(seconds=10))

                age = self.session.get_expiry_age(modification=modification)
                self.assertEqual(age, 10)

        def test_custom_expiry_datetime(self):
            modification = datetime.now()

            self.session.set_expiry(modification + timedelta(seconds=10))

            date = self.session.get_expiry_date(modification=modification)
            self.assertEqual(date, modification + timedelta(seconds=10))

            age = self.session.get_expiry_age(modification=modification)
            self.assertEqual(age, 10)

        def test_custom_expiry_reset(self):
            self.session.set_expiry(None)
            self.session.set_expiry(10)
            self.session.set_expiry(None)
            self.assertEqual(self.session.get_expiry_age(), SESSION_COOKIE_AGE)

        def test_decode(self):
            # Ensure we can decode what we encode
            data = {'a test key': 'a test value'}
            encoded = self.session.encode(data)
            self.assertEqual(self.session.decode(encoded), data)

        def test_cycle_data(self):
            session_id = str(uuid.uuid4())
            session = self.backend(session_id)
            session['name'] = 'test_cycle_data'
            session['session_id'] = session_id
            session.save()

            # before cycling data
            # check data has been saved
            self.assertEquals(session.items(),
                              self.backend(session_id).items())

            # saving items before they are cleared
            pre_session_items = session.items()

            new_session_id = session.cycle_data()

            # check a new session id was generated
            self.assertNotEqual(session_id, new_session_id)

            new_session = self.backend(new_session_id)
            # sanity check
            self.assertEquals(new_session['session_id'], session_id)
            # check data is the same
            self.assertEquals(new_session.items(), pre_session_items)

            # check that data has been removed in the previous version
            previous_session = self.backend(session_id)
            del previous_session['_session_expiry']
            self.assertEquals(list(previous_session.items()), [])


class TestWithFileObject(SessionTest.SessionTestsMixin):

    def backend(self, session_key=None):
        return SessionStore(session_key, FilesystemStore("./session_data_test"))



__author__ = 'Mwaaas'
__email__ = 'francismwangi152@gmail.com'
__phone_number__ = '+254702729654'


from marshmallow import fields

fields.Field.default_error_messages["required"] = 'This field is required.'



from copy import deepcopy
from ussd.core import UssdEngine
from ussd.store.journey_store import DummyStore, DynamoDb, YamlJourneyStore
from unittest import TestCase
from marshmallow.exceptions import ValidationError


class TestDriverStore:
    class BaseDriverStoreTestCase(TestCase):
        maxDiff = None

        def setUp(self):
            # setup driver instance
            self.driver = self.setup_driver()

        def tearDown(self):
            self.driver.flush()

        def test_creating_invalid_journey(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
            }

            self.assertRaises(ValidationError,
                              self.driver.save,
                              name="journey_a",
                              journey=sample_journey,
                              version="0.0.2")

        def test_creating_journey(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
                "end_screen": {
                    "type": "quit_screen",
                    "text": "end screen"
                }
            }

            # Creating journey_a
            sample_journey_one = deepcopy(sample_journey)

            sample_journey_two = deepcopy(sample_journey_one)
            sample_journey_two['end_screen']['text'] = "end screen of sample journey two"

            sample_journey_three = deepcopy(sample_journey)
            sample_journey_three['end_screen']['text'] = "end screen of sample journey three"

            self.driver.save(name="journey_a", journey=sample_journey_one, version="0.0.1")
            self.driver.save(name="journey_a", journey=sample_journey_two, version="0.0.2")
            self.driver.save(name="journey_a", journey=sample_journey_three, version="0.0.3")

            # Creating journey_b
            sample_journey_b_one = deepcopy(sample_journey)
            sample_journey_b_one['end_screen']['text'] = "end screen of sample journey_b one"

            sample_journey_b_two = deepcopy(sample_journey_b_one)
            sample_journey_b_two['end_screen']['text'] = "end screen of sample journey_b two"

            sample_journey_b_three = deepcopy(sample_journey_b_one)
            sample_journey_b_three['end_screen']['text'] = "end screen of sample journey_b three"

            self.driver.save(name="journey_b", journey=sample_journey_b_one, version="0.0.1")
            self.driver.save(name="journey_b", journey=sample_journey_b_two, version="0.0.2")
            self.driver.save(name="journey_b", journey=sample_journey_b_three, version="0.0.3")

            # test getting the journey a
            a = sample_journey_two
            b = self.driver.get('journey_a', '0.0.2')
            self.assertDictEqual(sample_journey_two, self.driver.get('journey_a', '0.0.2'))
            self.assertEqual(sample_journey_three, self.driver.get('journey_a', '0.0.3'))
            self.assertEqual(sample_journey_one, self.driver.get('journey_a', '0.0.1'))

            # test getting the journey b
            self.assertEqual(sample_journey_b_two, self.driver.get('journey_b', '0.0.2'))
            self.assertEqual(sample_journey_b_three, self.driver.get('journey_b', '0.0.3'))
            self.assertEqual(sample_journey_b_one, self.driver.get('journey_b', '0.0.1'))

            # test that you cannot save/update with the same version
            self.assertRaises(ValidationError, self.driver.save, name="journey_a",
                              journey=sample_journey_two, version="0.0.3")

            # test getting all the journeys
            self.assertDictEqual(
                {"0.0.1": sample_journey_one, "0.0.2": sample_journey_two, "0.0.3": sample_journey_three},
                self.driver.get_all_journey_version('journey_a'))

            # test getting the latest journey
            self.assertEqual(sample_journey_three, self.driver.get('journey_a'))

            # test getting part of screen alone
            self.assertEqual(sample_journey_three['end_screen'],
                             self.driver.get('journey_a', screen_name='end_screen'))
            self.assertEqual(sample_journey_two['end_screen'],
                             self.driver.get('journey_a', version="0.0.2", screen_name='end_screen'))

            # sanity check
            assert sample_journey_two['end_screen'] != sample_journey['end_screen']

            # testing deleting
            self.driver.delete('journey_a', version="0.0.1")
            self.assertIsNone(self.driver.get('journey_a', version="0.0.1", propagate_error=False))

            # test getting journey that doesn't exists throws an error
            self.assertRaises(ValidationError, self.driver.get, 'journey_a', version="0.0.1")

            # check that it hasn't deleted anything
            self.assertEqual(len(self.driver.get_all_journey_version('journey_a')), 2)

            # deleting all journeys including the versions
            self.driver.delete('journey_a')
            self.assertEqual(len(self.driver.get_all_journey_version('journey_a')), 0)

        def test_saving_journeys_thats_still_in_edit_mode(self):
            sample_journey = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
                "end_screen": {
                    "type": "quit_screen",
                    "text": "end screen"
                }
            }

            self.driver.save(name="journey_a", journey=sample_journey, version="0.0.1")

            self.assertEqual(sample_journey, self.driver.get('journey_a', '0.0.1'))

            sample_journey_edit_mode = {
                "initial_screen": {
                    "type": "initial_screen",
                    "next_screen": "end_screen",
                    "default_language": "en"
                },
            }

            # confirm its invalid
            is_valid, errors = UssdEngine.validate_ussd_journey(sample_journey_edit_mode)
            self.assertFalse(is_valid)

            # but we can still save it under edit mode
            self.driver.save(name='journey_a', journey=sample_journey_edit_mode, edit_mode=True)

            # stills get the valid one that was saved.
            self.assertEqual(sample_journey, self.driver.get('journey_a'))

            # getting the one use edit mode
            self.assertEqual(sample_journey_edit_mode, self.driver.get('journey_a', edit_mode=True))

            sample_journey_edit_mode['invalid_screen'] = "invalid_screen"

            # and we can still save it again
            self.driver.save(name='journey_a', journey=sample_journey_edit_mode, edit_mode=True)
            self.assertEqual(sample_journey_edit_mode, self.driver.get('journey_a', edit_mode=True))

        def test_crud_for_multiple_users(self):
            driver_with_user_1 = self.setup_driver(user="user_1")

            driver_with_user_2 = self.setup_driver(user="user_2")

            user_1 = {
                "journey_a": {
                    "0.0.1": {
                        "initial_screen": {
                            "type": "initial_screen",
                            "next_screen": "end_screen",
                            "default_language": "en"
                        },
                        "end_screen": {
                            "type": "quit_screen",
                            "text": "end screen for user 1 journey a 1"
                        }
                    },
                    "0.0.2": {
                        "initial_screen": {
                            "type": "initial_screen",
                            "next_screen": "end_screen",
                            "default_language": "en"
                        },
                        "end_screen": {
                            "type": "quit_screen",
                            "text": "end screen for user 1 journey a version 1"
                        }
                    }
                },
                "journey_b": {
                    "0.0.1": {
                        "initial_screen": {
                            "type": "initial_screen",
                            "next_screen": "end_screen",
                            "default_language": "en"
                        },
                        "end_screen": {
                            "type": "quit_screen",
                            "text": "end screen for user 1 journey b version 1"
                        }
                    }
                }
            }

            # create journey for user one
            driver_with_user_1.save(name="journey_a", version="0.0.1",
                                    journey=user_1["journey_a"]["0.0.1"])

            # sanity check journey was created
            self.assertEqual(
                driver_with_user_1.get("journey_a", version="0.0.1"),
                user_1["journey_a"]["0.0.1"])

            # check second driver can't access the journey with different user
            self.assertRaises(ValidationError, driver_with_user_2.get, "journey_a", version="0.0.1")

            # create another journey with a different version
            driver_with_user_1.save(name="journey_a", version="0.0.2",
                                    journey=user_1["journey_a"]["0.0.2"])

            # create another journey
            driver_with_user_1.save(name="journey_b", version="0.0.1",
                                    journey=user_1["journey_b"]["0.0.1"])

            # test getting all the journeys
            self.assertDictEqual(
                user_1["journey_a"],
                driver_with_user_1.get_all_journey_version('journey_a'))

            # test getting all the journeys returns for user one only
            self.assertDictEqual(
                user_1,
                driver_with_user_1.all())

            # now create journey for user 2
            user_2 = {
                "journey_a": {
                    "0.0.1": {
                        "initial_screen": {
                            "type": "initial_screen",
                            "next_screen": "end_screen",
                            "default_language": "en"
                        },
                        "end_screen": {
                            "type": "quit_screen",
                            "text": "end screen for user 2 journey a 1"
                        }
                    }
                },
                "journey_user_2": {
                    "0.0.1": {
                        "initial_screen": {
                            "type": "initial_screen",
                            "next_screen": "end_screen",
                            "default_language": "en"
                        },
                        "end_screen": {
                            "type": "quit_screen",
                            "text": "end screen for user 2 journey a 1"
                        }
                    }
                }
            }

            driver_with_user_2.save(name="journey_a", version="0.0.1",
                                    journey=user_2["journey_a"]['0.0.1'])

            self.assertEqual(
                driver_with_user_2.get("journey_a", version="0.0.1"),
                user_2["journey_a"]["0.0.1"]
            )

            # create the other journey
            driver_with_user_2.save(name="journey_user_2", version="0.0.1",
                                    journey=user_2["journey_user_2"]['0.0.1'])

            # check journey_a for user one is not equal as journey for user 2
            self.assertNotEqual(
                driver_with_user_1.get("journey_a", version="0.0.1"),
                driver_with_user_2.get("journey_a", version="0.0.1"),
            )

            # test getting all the journeys
            self.assertDictEqual(
                user_2["journey_a"],
                driver_with_user_2.get_all_journey_version('journey_a'))

            # test getting all the journeys returns for user one only
            self.assertDictEqual(
                user_2,
                driver_with_user_2.all())


class TestDummyStore(TestDriverStore.BaseDriverStoreTestCase):

    @staticmethod
    def setup_driver(user="default") -> DummyStore:
        return DummyStore.DummyStore(user=user)


class TestDynamodb(TestDriverStore.BaseDriverStoreTestCase):
    table_name = "test_dynamodb_journey_store"

    @classmethod
    def setUpClass(cls) -> None:
        DynamoDb.create_table(table_name=cls.table_name)

    @classmethod
    def tearDownClass(cls) -> None:
        DynamoDb.delete_table(table_name=cls.table_name)

    def setup_driver(self, user="default") -> DynamoDb:
        return DynamoDb.DynamoDb(self.table_name, "http://dynamodb:8000", user=user)


class TestYamlJourneyStore(TestDriverStore.BaseDriverStoreTestCase):

    @staticmethod
    def setup_driver(user="default") -> YamlJourneyStore:
        return YamlJourneyStore.YamlJourneyStore(user=user)



import datetime
from decimal import Decimal


class UnicodeDecodeError(UnicodeDecodeError):
    def __init__(self, obj, *args):
        self.obj = obj
        super().__init__(*args)

    def __str__(self):
        return '%s. You passed in %r (%s)' % (super().__str__(), self.obj, type(self.obj))


_PROTECTED_TYPES = (
    type(None), int, float, Decimal, datetime.datetime, datetime.date, datetime.time,
)


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to
    force_text(strings_only=True).
    """
    return isinstance(obj, _PROTECTED_TYPES)


def force_text(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Similar to smart_text, except that lazy instances are resolved to
    strings, rather than kept as lazy objects.

    If strings_only is True, don't convert (some) non-string-like objects.
    """
    # Handle the common case first for performance reasons.
    if issubclass(type(s), str):
        return s
    if strings_only and is_protected_type(s):
        return s
    try:
        if isinstance(s, bytes):
            s = str(s, encoding, errors)
        else:
            s = str(s)
    except UnicodeDecodeError as e:
        raise UnicodeDecodeError(s, *e.args)
    return s



import inspect


def method_has_no_args(meth):
    """Return True if a method only accepts 'self'."""
    count = len([
        p for p in inspect.signature(meth).parameters.values()
        if p.kind == p.POSITIONAL_OR_KEYWORD
    ])
    return count == 0 if inspect.ismethod(meth) else count == 1



import importlib
import os
from datetime import datetime

import yaml
from jinja2 import Template
from yaml.constructor import ConstructorError

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import SafeLoader as Loader

date_format = "%Y-%m-%d %H:%M:%S.%f"


def str_to_class(import_path):
    module_name, class_name = import_path.rsplit(".", 1)
    try:
        module_ = importlib.import_module(module_name)
        try:
            class_ = getattr(module_, class_name)
        except AttributeError:
            raise Exception('Class does not exist')
    except ImportError:
        raise Exception('Module does not exist')
    return class_


def datetime_to_string(date_obj: datetime):
    return date_obj.strftime(date_format)


def string_to_datetime(date_str_obj: str):
    return datetime.strptime(date_str_obj, date_format)


def get_text(text):
    if type(text) is str:
        return text
    return text['en']


def extract_file(file_name, parent_file_path):
    # first try the raw path

    if os.path.isfile(file_name):
        file_path = file_name
    else:
        file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(parent_file_path),
                file_name
            )
        )
        if not os.path.isfile(file_path):
            raise FileNotFoundError(file_name)

    with open(file_path, 'r') as f:
        return yaml.load(f)


def include(loader, node):
    if isinstance(node, yaml.ScalarNode):
        return extract_file(loader.construct_scalar(node), node.start_mark.name)

    elif isinstance(node, yaml.SequenceNode):
        data = {}
        for filename in loader.construct_sequence(node):
            file_path = Template(filename).render(os.environ)
            data.update(extract_file(os.path.abspath(file_path)))

        return data

    else:
        raise ConstructorError("Error:: unrecognised node type in !include statement")



class YamlToGo:
    def __init__(self, file):
        with open(file, 'r') as f:
            data = yaml.load(f)

        included_files = data.get('includes')
        if included_files:
            data.pop('includes', None)
            data.update(included_files)
        self.yaml = data

        self.count = 0

    def get_model_data(self):
        data = {}
        links = []
        for index, value in enumerate(self.yaml):
            data[value] = self.convert_screen(value, index)

        for index, value in enumerate(self.yaml):
            links.extend(self.get_links(value, data))

        data_array = [data.pop('initial_screen')]
        data_array.extend(list(data.values()))

        return {'data': data_array, 'links': links}

    def convert_screen(self, screen, key):
        _data = {}
        name = screen
        screen = self.yaml[screen]
        screen_type = screen['type']
        _data['key'] = key
        _data['title'] = name

        if screen_type == 'initial_screen':
            _data['items'] = [
                {'index': 1, 'portName': 'out1', 'text': 'Initial Screen'}
            ]
        elif screen_type == 'input_screen':
            _data['items'] = [
                {'index': '1', 'portName': 'out1', 'text': get_text(screen['text'])}
            ]
        elif screen_type.endswith('http_screen'):
            _data['items'] = [
                {'index': '1', 'portName': 'out1', 'text': 'Http Request Screen'}
            ]

        elif screen_type == 'menu_screen':
            _data['items'] = []
            for index, value in enumerate(screen['options']):
                _data['items'].append({
                    'index': index + 1,
                    'text': get_text(value['text']),
                    'portName': 'out' + str(index)
                })

        elif screen_type == 'router_screen':
            _data['items'] = []

            for index, value in enumerate(screen['router_options']):
                _data['items'].append({
                    'index': index + 1,
                    'text': value['expression'],
                    'portName': 'out' + str(index)
                })

            if 'default_next_screen' in screen:
                _data['items'].append({
                    'index': len(screen['router_options']) + 1,
                    'text': 'Default Route',
                    'portName': 'outDefault'
                })

        elif screen_type == 'quit_screen':
            _data['items'] = [
                {'index': 1, 'portName': 'out1', 'text': get_text(screen['text'])}
            ]

        return _data

    def get_links(self, screen, data):
        _data = []
        name = screen
        screen = self.yaml[screen]
        screen_type = screen['type']

        if screen_type == 'initial_screen':
            _data = [
                {
                    'from': data[name]['key'],
                    'to': data[screen['next_screen']]['key'],
                    'fromPort': 'out1'
                }
            ]
        elif screen_type == 'input_screen':
            _data = [
                {
                    'from': data[name]['key'],
                    'to': data[screen['next_screen']]['key'],
                    'fromPort': 'out1'
                }
            ]
        elif screen_type.endswith('http_screen'):
            _data = [
                {
                    'from': data[name]['key'],
                    'to': data[screen['next_screen']]['key'],
                    'fromPort': 'out1'
                }
            ]

        elif screen_type == 'menu_screen':
            for index, value in enumerate(screen['options']):
                _data.append({
                    'from': data[name]['key'],
                    'to': data[value['next_screen']]['key'],
                    'fromPort': 'out' + str(index)
                })

        elif screen_type == 'router_screen':
            if 'default_next_screen' in screen:
                _data.append({
                    'from': data[name]['key'],
                    'to': data[screen['default_next_screen']]['key'],
                    'fromPort': 'outDefault'
                })
            for index, value in enumerate(screen['router_options']):
                _data.append({
                    'from': data[name]['key'],
                    'to': data[value['next_screen']]['key'],
                    'fromPort': 'out' + str(index)
                })

        elif screen_type == 'quit_screen':
            pass

        return _data



from ussd.store.journey_store import JourneyStore
import os
import yaml
from jinja2 import Template, Environment
from configure import Configuration
from ussd import utilities
from yaml.constructor import ConstructorError
import staticconf
from shutil import rmtree


def extract_file(file_name, parent_file_path):
    # first try the raw path

    if os.path.isfile(file_name):
        file_path = file_name
    else:
        file_path = os.path.abspath(
            os.path.join(
                os.path.dirname(parent_file_path),
                file_name
            )
        )
        if not os.path.isfile(file_path):
            raise FileNotFoundError(file_name)

    with open(file_path, 'r') as f:
        return yaml.load(f)


def include(loader, node):
    if isinstance(node, yaml.ScalarNode):
        return extract_file(loader.construct_scalar(node), node.start_mark.name)

    elif isinstance(node, yaml.SequenceNode):
        data = {}
        for filename in loader.construct_sequence(node):
            file_path = Template(filename).render(os.environ)
            data.update(extract_file(os.path.abspath(file_path)))

        return data

    else:
        raise ConstructorError("Error:: unrecognised node type in !include statement")


def load_dict_from_yaml(file_path) -> dict:
    file_path = Template(file_path).render(os.environ)
    return Configuration.from_file(
        os.path.abspath(file_path),
        multi_constructors={'!include': include},
        configure=False
    )


def load_yaml(file_path) -> dict:
    if file_path not in staticconf.config.configuration_namespaces:
        yaml_dict = load_dict_from_yaml(file_path)
        staticconf.DictConfiguration(
            yaml_dict,
            namespace=file_path,
            flatten=False)
    return staticconf.config.configuration_namespaces[file_path].configuration_values


class YamlJourneyStore(JourneyStore):
    """
    Loader used for loading and using journeys in a yaml file
    """

    def __init__(self, user="default", journey_directory="./.journeys"):
        self.journey_directory = os.path.join(
            os.path.abspath(journey_directory),
            user)

        if not os.path.isdir(self.journey_directory):
            os.makedirs(self.journey_directory)

    def _get_or_create_directory(self, name):
        directory = self._get_directory(name)
        if not os.path.isdir(directory):
            os.makedirs(directory)
        return directory

    def _get_directory(self, name):
        return os.path.join(self.journey_directory, "{0}".format(name))

    def _get_file_path(self, name, version):
        return os.path.join(self._get_directory(name),
                            "{0}.yml".format(version))

    def _get(self, name, version, screen_name, **kwargs):
        if version is None:
            files = os.listdir(self._get_directory(name))
            files.sort()
            version = files[-1]
            if version.replace(".yml", "") == self.edit_mode_version:
                version = files[-2]
            version = version.replace(".yml", "")

        file_path = self._get_file_path(name, version)
        if not os.path.isfile(file_path):
            return None

        journey = load_yaml(file_path)
        if screen_name:
            return journey[screen_name]
        return journey

    def _get_all_journey_version(self, name):
        directory = self._get_directory(name)

        if os.path.isdir(directory):
            files = os.listdir(directory)
        else:
            files = []

        results = {}
        for i in files:
            if i.endswith(".yml"):
                version = i.replace(".yml", "")
                results[version] = self._get(name, version, None)
        return results

    def _save(self, name, journey, version):
        file_path = self._get_file_path(name, version)
        if staticconf.config.configuration_namespaces.get(file_path):
            del staticconf.config.configuration_namespaces[file_path]
        self._get_or_create_directory(name)
        with open(file_path, 'w') as outfile:
            yaml.dump(journey, outfile, default_flow_style=False)

    def _all(self):
        results = {}
        for i in os.listdir(self.journey_directory):
            results[i] = self._get_all_journey_version(i)
        return results

    @staticmethod
    def _delete_folder(path):
        rmtree(path)

    def _delete(self, name, version=None):
        directory = self._get_directory(name)
        if version is None:
            self._delete_folder(directory)
        else:
            os.remove(self._get_file_path(name, version))

    def flush(self):
        self._delete_folder(self.journey_directory)



"""
Customer journey are stored in a document store.
Any engine that implements this interface can be integrated with journey store.
"""
import abc
import inspect
from copy import deepcopy
from ussd.utils.module_loading import import_string
import typing
from marshmallow.exceptions import ValidationError


class JourneyStore(object, metaclass=abc.ABCMeta):

    edit_mode_version = "edit_mode"

    @abc.abstractmethod
    def _get(self, name, version, screen_name, **kwargs):
        pass

    @abc.abstractmethod
    def _all(self):
        pass

    @abc.abstractmethod
    def _save(self, name, journey, version):
        pass

    @abc.abstractmethod
    def _delete(self, name, version=None):
        pass

    @abc.abstractmethod
    def _get_all_journey_version(self, name):
        pass

    @abc.abstractmethod
    def flush(self):
        pass

    def get(self, name: str, version=None, screen_name=None, edit_mode=False, propagate_error=True):
        if edit_mode:
            version = self.edit_mode_version
        results = self._get(name, version, screen_name)

        if propagate_error and results is None:
            raise ValidationError(
                "Journey with name {0} and version {1} does not exist".format(name, version))
        return results

    def all(self):
        return self._all()

    def get_all_journey_version(self, name):
        return self._get_all_journey_version(name)

    def save(self, name: str, journey: dict, version=None, edit_mode=False):

        # version and editor mode should not be false
        if not (version or edit_mode):
            raise ValidationError("version is required if its not in editor mode")

        if edit_mode:
            version = self.edit_mode_version

        # check if this version already exists
        if self.get(name, version, propagate_error=False) is not None:
            if not edit_mode:
                raise ValidationError("journey already exists")

        # validate if its not in editing mode.
        if not edit_mode:
            from ussd.core import UssdEngine
            is_valid, errors = UssdEngine.validate_ussd_journey(journey)
            if not is_valid:
                raise ValidationError(errors, "journey")

        # now create journey
        return self._save(name, journey, version)

    def delete(self, name, version=None):
        return self._delete(name, version)


class JourneyStoreApi(object):

    def __init__(self, 
                 driver: typing.Union[str, typing.Any] = "ussd.store.journey_store.YamlJourneyStore.YamlJourneyStore",
                 driver_config: typing.Dict = None):
        
        self.driver_config = {} \
            if driver_config is None else driver_config
        self.driver = driver

        if not inspect.isclass(driver):
            if not isinstance(self.driver, str):
                raise ValidationError(
                    "driver should either be an instance of "
                    "driver store or a string pointing to the store")
            else:
                self.driver = import_string(self.driver)

        if not inspect.isclass(self.driver):
            raise ValidationError("driver should be a class ")

        # initiate the driver
        self.driver = self.driver(**self.driver_config)

    journey_method_names = ("save", "delete", "get", "all", "flush")

    def __getattr__(self, name):
        if name in JourneyStoreApi.journey_method_names:
            return self._actual_function(getattr(self.driver, name))
        else:
            return super(JourneyStoreApi, self).__getattribute__(name)

    def _actual_function(self, journey_store_method):
        def wrapper(*args, **kwargs):
            errors = self._parameter_validation(journey_store_method, *args, **kwargs)
            if errors:
                raise ValidationError(errors)
            return journey_store_method(*args, **kwargs)
        return wrapper

    def handle_action(self, **kwargs):

        action = kwargs.pop('action', '')

        if not action:
            raise ValidationError("This field is required", "action")

        # For easy development should accept post and convert it to save
        if action in ('post', 'put'):
            action = 'save'
        if action == 'get' and len(kwargs) == 0:
            action = 'all'

        if action not in JourneyStoreApi.journey_method_names:
            raise ValidationError(
                "action '{0}' is not allowed, "
                "only this methods are allowed; {1}".format(
                    action,
                    ", ".join(JourneyStoreApi.journey_method_names),
                )
                , 'action')

        return getattr(self, action)(**kwargs)

    @staticmethod
    def _parameter_validation(method, *args, **kwargs) -> typing.Dict:

        assert inspect.isfunction(method) or inspect.ismethod(method)

        errors = {}

        parameters = list(inspect.signature(method).parameters.values())

        method_kwargs = deepcopy(kwargs)
        method_args = list(deepcopy(args))

        for i in parameters:
            if len(method_args) > 0:
                method_args = method_args[1:]
                continue
            if i.name not in method_kwargs.keys():
                if i.default == i.empty:
                    errors[i.name] = ["This field is required"]
            else:
                del method_kwargs[i.name]

        for i in method_kwargs:
            errors[i] = ["This field is not required"]

        return errors

    def get_journey_store(self) -> JourneyStore:
        return self.driver



from marshmallow import Schema, fields, validates, ValidationError, validate
from ussd.screens.fields import UssdTextField, NextUssdScreenField, WithDictField, WithItemField


class UssdBaseScreenSchema(Schema):
    type = fields.Str(required=True)

    @validates("type")
    def validate_type(self, value):
        # to avoid cyclic import
        from ussd.core import _registered_ussd_handlers
        if value not in _registered_ussd_handlers.keys():
            raise ValidationError("Invalid screen type not supported")
        return value


class UssdTextSchema(Schema):
    text = UssdTextField(required=True)


class UssdContentBaseSchema(UssdBaseScreenSchema, UssdTextSchema):
    pass


class NextUssdScreenSchema(Schema):
    next_screen = NextUssdScreenField(required=True)

    @validates("next_screen")
    def validate_next_screen(self, value):
        for next_screen_config in value:
            next_screen = next_screen_config.get('next_screen')
            if not next_screen:
                raise ValidationError(
                    {'next_screen': ['This field is required.']}
                )
            if next_screen not in self.context.keys():
                raise ValidationError(
                    "{screen} is missing in ussd journey".format(screen=next_screen)
                )
        return value


class MenuOptionSchema(UssdTextSchema, NextUssdScreenSchema):
    input_value = fields.Str(required=False, validate=validate.Length(max=5))
    input_display = fields.Str(required=False, validate=validate.Length(max=5))


class MenuSchema(Schema):
    options = fields.List(fields.Nested(MenuOptionSchema), required=True)


class WithItemSchema(Schema):
    with_items = WithItemField(required=False, default=None)


class WithDictSchema(Schema):
    with_dict = WithDictField(required=False, default=None)
